"""Thin RMAT setup wrapper.

Delegates shared installation behavior to root-level core.py.
"""

import argparse
import sys
from pathlib import Path


def build_hooks():
    """Return optional hook callbacks for model-specific custom steps."""
    return {}


def main() -> None:
    parser = argparse.ArgumentParser(description="Set up RMAT model artifacts and environment")
    parser.add_argument("--model", default="RMAT", help="Model name (e.g., RMAT)")
    parser.add_argument(
        "--download-only",
        action="store_true",
        help="Only download/checkpoint artifacts; skip venv/dependency installation",
    )
    parser.add_argument(
        "--workspace",
        default="",
        help="Workspace root where models/ artifacts are written",
    )
    args = parser.parse_args()

    metadata_dir = Path(__file__).resolve().parent
    package_root = metadata_dir.parent.parent
    workspace_root = Path(args.workspace).resolve() if args.workspace else package_root

    sys.path.insert(0, str(package_root))
    from core import run_setup_pipeline

    run_setup_pipeline(
        workspace_root=workspace_root,
        metadata_dir=metadata_dir,
        model_name=args.model,
        hooks=build_hooks(),
        stage="download_only" if args.download_only else "full",
    )


if __name__ == "__main__":
    main()



import argparse
import os
import sys
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import selfies as sf
import torch
import torch.nn.functional as F

SCRIPT_DIR = Path(__file__).resolve().parent
DEFAULT_HF_HOME = SCRIPT_DIR / ".cache" / "huggingface"
DEFAULT_TRANSFORMERS_CACHE = DEFAULT_HF_HOME / "transformers"

MODEL_VARIANTS: Dict[str, Dict[str, str]] = {
    "7b": {
        "repo_id": "zjunlp/MolGen-7b",
        "cache_subdir": "MolGen-7b",
    },
    "large": {
        "repo_id": "zjunlp/MolGen-large",
        "cache_subdir": "MolGen-large",
    },
    "large-opt": {
        "repo_id": "zjunlp/MolGen-large-opt",
        "cache_subdir": "MolGen-large-opt",
    },
}

# Ensure huggingface cache is model-local by default.
os.environ.setdefault("HF_HOME", str(DEFAULT_HF_HOME))
os.environ.setdefault("TRANSFORMERS_CACHE", str(DEFAULT_TRANSFORMERS_CACHE))
os.environ.setdefault("HF_HUB_CACHE", str(DEFAULT_HF_HOME / "hub"))
os.environ.setdefault("HF_HUB_DISABLE_SYMLINKS_WARNING", "1")
os.environ.setdefault("HF_HUB_DISABLE_IMPLICIT_TOKEN", "1")

try:
    from transformers import AutoConfig, AutoModel, AutoTokenizer  # type: ignore
except Exception as exc:
    raise ImportError(
        "Could not import transformers AutoConfig/AutoModel/AutoTokenizer. "
        "Ensure MOLGEN setup completed successfully."
    ) from exc


PROJECT_ROOT = Path(__file__).resolve().parent.parent.parent
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))

from embeddings_preprocess import (
    load_input_lines,
    preprocess_lines,
    resolve_output_paths,
    write_failures_csv,
    write_mapping_csv,
)


def _normalize_variant(variant: str) -> str:
    normalized = variant.strip().lower()
    if normalized not in MODEL_VARIANTS:
        allowed = ", ".join(sorted(MODEL_VARIANTS.keys()))
        raise ValueError(f"Unsupported MOLGEN variant: {variant}. Allowed: {allowed}")
    return normalized


def _resolve_device(device_arg: str) -> torch.device:
    normalized = device_arg.strip().lower()
    if normalized == "auto":
        return torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if normalized not in {"cpu", "cuda"}:
        raise ValueError(f"Unsupported device: {device_arg}")
    if normalized == "cuda" and not torch.cuda.is_available():
        print("CUDA requested but unavailable; falling back to CPU.", file=sys.stderr)
        return torch.device("cpu")
    return torch.device(normalized)


def _resolve_model_source(model_path: str, pretrained_name: str, variant: str) -> Tuple[str, bool]:
    explicit = model_path.strip()
    if explicit:
        candidate = Path(explicit).expanduser()
        if candidate.exists():
            return str(candidate.resolve()), True
        return explicit, False

    explicit_pretrained = pretrained_name.strip()
    if explicit_pretrained:
        candidate = Path(explicit_pretrained).expanduser()
        if candidate.exists():
            return str(candidate.resolve()), True
        return explicit_pretrained, False

    variant_cfg = MODEL_VARIANTS[variant]
    local_dir = DEFAULT_HF_HOME / variant_cfg["cache_subdir"]
    if local_dir.exists():
        return str(local_dir.resolve()), True

    return variant_cfg["repo_id"], False


def _build_model(
    model_source: str,
    revision: str,
    trust_remote_code: bool,
    local_files_only: bool,
    device: torch.device,
):
    kwargs = {
        "trust_remote_code": trust_remote_code,
        "local_files_only": local_files_only,
    }
    if revision:
        kwargs["revision"] = revision

    config = AutoConfig.from_pretrained(model_source, **kwargs)
    tokenizer = AutoTokenizer.from_pretrained(model_source, **kwargs)
    model = AutoModel.from_pretrained(model_source, **kwargs).to(device).eval()

    if tokenizer.pad_token is None and tokenizer.eos_token is not None:
        tokenizer.pad_token = tokenizer.eos_token

    return config, tokenizer, model


def _to_molgen_text(canonical_smiles: str) -> str:
    selfies_text = sf.encoder(canonical_smiles)
    return selfies_text.replace("][", "] [")


@torch.inference_mode()
def _encode_batches(
    model_inputs: List[str],
    config,
    model,
    tokenizer,
    device: torch.device,
    batch_size: int,
    max_length: int,
    normalize: bool,
) -> np.ndarray:
    chunks = []

    for start in range(0, len(model_inputs), batch_size):
        batch_texts = model_inputs[start : start + batch_size]
        encoded = tokenizer(
            batch_texts,
            padding=True,
            truncation=True,
            max_length=max_length,
            return_tensors="pt",
        )
        encoded = {key: value.to(device) for key, value in encoded.items()}

        if bool(getattr(config, "is_encoder_decoder", False)) and hasattr(model, "get_encoder"):
            encoder_outputs = model.get_encoder()(
                input_ids=encoded["input_ids"],
                attention_mask=encoded.get("attention_mask"),
            )
            last_hidden = encoder_outputs.last_hidden_state
        else:
            outputs = model(
                input_ids=encoded["input_ids"],
                attention_mask=encoded.get("attention_mask"),
            )
            last_hidden = getattr(outputs, "last_hidden_state", None)
            if last_hidden is None:
                if isinstance(outputs, (tuple, list)) and outputs:
                    last_hidden = outputs[0]
                else:
                    raise RuntimeError("MOLGEN forward outputs did not include last_hidden_state.")

        mask = encoded.get("attention_mask")
        if mask is None:
            mask = torch.ones(last_hidden.shape[:2], dtype=torch.long, device=device)
        mask = mask.unsqueeze(-1).type_as(last_hidden)
        emb = (last_hidden * mask).sum(1) / mask.sum(1).clamp(min=1e-6)

        if normalize:
            emb = F.normalize(emb, dim=-1)

        chunks.append(emb.detach().cpu().numpy().astype(np.float32, copy=False))

    if not chunks:
        raise RuntimeError("No embeddings were produced by MOLGEN.")
    return np.concatenate(chunks, axis=0)


def main() -> None:
    parser = argparse.ArgumentParser(description="Encode molecules with MOLGEN and save embeddings")
    parser.add_argument("--file", required=True, help="Path to input .txt file")
    parser.add_argument(
        "--input-format",
        default="selfies",
        choices=["selfies", "smiles"],
        help="Format of rows in input file",
    )
    parser.add_argument(
        "--model-input-type",
        default="smiles",
        choices=["selfies", "smiles"],
        help="Expected model input type from metadata",
    )
    parser.add_argument("--output-dir", default=".", help="Directory to write outputs")
    parser.add_argument("--output-stem", default="embeddings", help="Base output filename stem")
    parser.add_argument(
        "--variant",
        default="7b",
        choices=sorted(MODEL_VARIANTS.keys()),
        help="MolGen checkpoint variant",
    )
    parser.add_argument(
        "--pretrained-name",
        default="",
        help="Override model source (HF repo id or local model path)",
    )
    parser.add_argument(
        "--model-path",
        default="",
        help="Optional explicit local model directory (highest priority)",
    )
    parser.add_argument(
        "--revision",
        default="",
        help="Optional model revision override",
    )
    parser.add_argument("--batch-size", type=int, default=8)
    parser.add_argument("--max-length", type=int, default=256)
    parser.add_argument(
        "--device",
        default="auto",
        choices=["auto", "cpu", "cuda"],
        help="Inference device",
    )
    parser.add_argument(
        "--trust-remote-code",
        dest="trust_remote_code",
        action="store_true",
        help="Allow custom model code from the Hugging Face repository",
    )
    parser.add_argument(
        "--no-trust-remote-code",
        dest="trust_remote_code",
        action="store_false",
        help="Disable remote model code loading",
    )
    parser.add_argument(
        "--no-normalize",
        dest="normalize",
        action="store_false",
        help="Disable L2-normalization before saving",
    )
    parser.set_defaults(normalize=True, trust_remote_code=True)
    args = parser.parse_args()

    input_path = Path(args.file).resolve()
    if not input_path.exists():
        print(f"Input file not found: {input_path}", file=sys.stderr)
        sys.exit(1)

    output_dir = Path(args.output_dir).resolve()
    npy_path, map_path, fail_path = resolve_output_paths(output_dir, args.output_stem)

    lines = load_input_lines(input_path)
    records, failures = preprocess_lines(lines, args.input_format, args.model_input_type)
    if not records:
        print("No valid molecules after circularity validation. Aborting.", file=sys.stderr)
        if failures:
            print(f"Failures: {len(failures)} (see {fail_path})", file=sys.stderr)
            write_failures_csv(fail_path, failures)
        sys.exit(1)

    molgen_inputs: List[str] = []
    filtered_records = []
    for record in records:
        try:
            molgen_inputs.append(_to_molgen_text(record["smiles"]))
            filtered_records.append(record)
        except Exception:
            failures.append(
                {
                    "original_line": record["original_line"],
                    "input_text": record["input_text"],
                    "reason": "selfies_encoding_failed",
                }
            )

    if not molgen_inputs:
        print("No valid molecules after MolGen token preparation. Aborting.", file=sys.stderr)
        if failures:
            write_failures_csv(fail_path, failures)
        sys.exit(1)

    variant = _normalize_variant(args.variant)
    model_source, local_files_only = _resolve_model_source(args.model_path, args.pretrained_name, variant)
    device = _resolve_device(args.device)

    config, tokenizer, model = _build_model(
        model_source=model_source,
        revision=args.revision.strip(),
        trust_remote_code=args.trust_remote_code,
        local_files_only=local_files_only,
        device=device,
    )

    embeddings = _encode_batches(
        model_inputs=molgen_inputs,
        config=config,
        model=model,
        tokenizer=tokenizer,
        device=device,
        batch_size=args.batch_size,
        max_length=args.max_length,
        normalize=args.normalize,
    )

    np.save(npy_path, embeddings)
    write_mapping_csv(map_path, filtered_records)
    if failures:
        write_failures_csv(fail_path, failures)

    print(f"Saved embeddings to {npy_path} with shape {embeddings.shape}.")
    print(f"Saved row mapping to {map_path}.")
    print(f"MolGen variant: {variant}")
    print(f"Model source: {model_source} (local_files_only={local_files_only})")
    if args.revision.strip():
        print(f"Revision: {args.revision.strip()}")
    if failures:
        print(f"Filtered {len(failures)} invalid rows. See {fail_path}.")


if __name__ == "__main__":
    main()


import argparse
import os
import sys
from pathlib import Path
from typing import List

import numpy as np
import torch
import torch.nn.functional as F

SCRIPT_DIR = Path(__file__).resolve().parent
DEFAULT_HM_CACHE = SCRIPT_DIR / ".cache" / "huggingmolecules"
# huggingmolecules snapshots cache path at import time, so set it before imports.
os.environ.setdefault("HUGGINGMOLECULES_CACHE", str(DEFAULT_HM_CACHE))

try:
    from huggingmolecules import GroverFeaturizer, GroverModel  # type: ignore
except Exception as exc:
    raise ImportError(
        "Could not import GROVER classes from huggingmolecules. "
        "Ensure GROVER setup completed successfully."
    ) from exc


PROJECT_ROOT = Path(__file__).resolve().parent.parent.parent
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))

from embeddings_preprocess import (
    load_input_lines,
    preprocess_lines,
    resolve_output_paths,
    write_failures_csv,
    write_mapping_csv,
)


def _resolve_device(device_arg: str) -> torch.device:
    normalized = device_arg.strip().lower()
    if normalized == "auto":
        return torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if normalized not in {"cpu", "cuda"}:
        raise ValueError(f"Unsupported device: {device_arg}")
    if normalized == "cuda" and not torch.cuda.is_available():
        print("CUDA requested but unavailable; falling back to CPU.", file=sys.stderr)
        return torch.device("cpu")
    return torch.device(normalized)


def _resolve_pretrained_name(variant: str, explicit_name: str) -> str:
    if explicit_name:
        return explicit_name
    normalized = variant.strip().lower()
    if normalized not in {"base", "large"}:
        raise ValueError(f"Unsupported GROVER variant: {variant}")
    return f"grover_{normalized}"


def _build_model(pretrained_name: str, device: torch.device):
    featurizer = GroverFeaturizer.from_pretrained(pretrained_name)
    model = GroverModel.from_pretrained(pretrained_name).to(device).eval()
    return featurizer, model


@torch.inference_mode()
def _encode_smiles_batches(
    smiles_inputs: List[str],
    model,
    featurizer,
    device: torch.device,
    batch_size: int,
    use_atoms: bool,
    use_bonds: bool,
    normalize: bool,
) -> np.ndarray:
    if not use_atoms and not use_bonds:
        raise ValueError("At least one GROVER view must be enabled: atoms and/or bonds.")

    captured = {"atoms": None, "bonds": None}

    def _hook_atoms(_module, inputs, _output):
        captured["atoms"] = inputs[0].detach()

    def _hook_bonds(_module, inputs, _output):
        captured["bonds"] = inputs[0].detach()

    atom_head = getattr(model, "mol_atom_from_atom_ffn", None)
    bond_head = getattr(model, "mol_atom_from_bond_ffn", None)
    if atom_head is None or bond_head is None:
        raise RuntimeError("GROVER model is missing expected FFN heads for atom/bond outputs.")

    hook_atoms = atom_head.register_forward_hook(_hook_atoms)
    hook_bonds = bond_head.register_forward_hook(_hook_bonds)

    chunks = []
    try:
        for start in range(0, len(smiles_inputs), batch_size):
            batch_smiles = smiles_inputs[start : start + batch_size]
            batch = featurizer(batch_smiles)
            if hasattr(batch, "to"):
                batch = batch.to(device)

            captured["atoms"] = None
            captured["bonds"] = None
            _ = model(batch)

            parts = []
            if use_atoms:
                if captured["atoms"] is None:
                    raise RuntimeError("Failed to capture GROVER atom embeddings.")
                parts.append(captured["atoms"])
            if use_bonds:
                if captured["bonds"] is None:
                    raise RuntimeError("Failed to capture GROVER bond embeddings.")
                parts.append(captured["bonds"])

            emb = parts[0] if len(parts) == 1 else torch.cat(parts, dim=1)
            if normalize:
                emb = F.normalize(emb, dim=-1)
            chunks.append(emb.cpu().numpy().astype(np.float32, copy=False))
    finally:
        hook_atoms.remove()
        hook_bonds.remove()

    if not chunks:
        raise RuntimeError("No embeddings were produced by GROVER.")
    return np.concatenate(chunks, axis=0)


def main() -> None:
    parser = argparse.ArgumentParser(description="Encode molecules with GROVER and save embeddings")
    parser.add_argument("--file", required=True, help="Path to input .txt file")
    parser.add_argument(
        "--input-format",
        default="selfies",
        choices=["selfies", "smiles"],
        help="Format of rows in input file",
    )
    parser.add_argument(
        "--model-input-type",
        default="selfies",
        choices=["selfies", "smiles"],
        help="Expected model input type from metadata",
    )
    parser.add_argument("--output-dir", default=".", help="Directory to write outputs")
    parser.add_argument("--output-stem", default="embeddings", help="Base output filename stem")
    parser.add_argument(
        "--variant",
        default="base",
        choices=["base", "large"],
        help="GROVER checkpoint variant",
    )
    parser.add_argument(
        "--pretrained-name",
        default="",
        help="Override checkpoint key (for example, grover_base or grover_large)",
    )
    parser.add_argument("--use-atoms", action="store_true", help="Include atom-view embeddings")
    parser.add_argument("--use-bonds", action="store_true", help="Include bond-view embeddings")
    parser.add_argument("--batch-size", type=int, default=32)
    parser.add_argument(
        "--device",
        default="auto",
        choices=["auto", "cpu", "cuda"],
        help="Inference device",
    )
    parser.add_argument(
        "--no-normalize",
        dest="normalize",
        action="store_false",
        help="Disable L2-normalization before saving",
    )
    parser.set_defaults(normalize=True)
    args = parser.parse_args()

    # Conservative default keeps behavior stable if no view flag is passed.
    use_atoms = args.use_atoms or not args.use_bonds
    use_bonds = args.use_bonds

    input_path = Path(args.file).resolve()
    if not input_path.exists():
        print(f"Input file not found: {input_path}", file=sys.stderr)
        sys.exit(1)

    output_dir = Path(args.output_dir).resolve()
    npy_path, map_path, fail_path = resolve_output_paths(output_dir, args.output_stem)

    lines = load_input_lines(input_path)
    records, failures = preprocess_lines(lines, args.input_format, args.model_input_type)
    if not records:
        print("No valid molecules after circularity validation. Aborting.", file=sys.stderr)
        if failures:
            print(f"Failures: {len(failures)} (see {fail_path})", file=sys.stderr)
            write_failures_csv(fail_path, failures)
        sys.exit(1)

    # GROVER featurizer consumes SMILES; preprocess_lines already provides canonical SMILES.
    smiles_inputs = [record["smiles"] for record in records]

    pretrained_name = _resolve_pretrained_name(args.variant, args.pretrained_name)
    device = _resolve_device(args.device)
    featurizer, model = _build_model(pretrained_name, device)
    embeddings = _encode_smiles_batches(
        smiles_inputs=smiles_inputs,
        model=model,
        featurizer=featurizer,
        device=device,
        batch_size=args.batch_size,
        use_atoms=use_atoms,
        use_bonds=use_bonds,
        normalize=args.normalize,
    )

    np.save(npy_path, embeddings)
    write_mapping_csv(map_path, records)
    if failures:
        write_failures_csv(fail_path, failures)

    print(f"Saved embeddings to {npy_path} with shape {embeddings.shape}.")
    print(f"Saved row mapping to {map_path}.")
    print(f"GROVER options: pretrained={pretrained_name}, atoms={use_atoms}, bonds={use_bonds}")
    if failures:
        print(f"Filtered {len(failures)} invalid rows. See {fail_path}.")


if __name__ == "__main__":
    main()



import numpy as np
from sklearn.gaussian_process.kernels import RBF, Matern
from sklearn.gaussian_process.kernels import ConstantKernel as C
from sklearn.gaussian_process.kernels import Kernel, Hyperparameter

class LaplacianKernel(Kernel):
    """A Laplacian kernel compatible with scikit-learn's GaussianProcessRegressor."""
    def __init__(self, length_scale=1.0, length_scale_bounds=(1e-5, 1e5)):
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds

    @property
    def hyperparameter_length_scale(self):
        return Hyperparameter("length_scale", "numeric", self.length_scale_bounds)

    def __call__(self, X, Y=None, eval_gradient=False):
        Y = X if Y is None else Y
        dists = np.sqrt(((X[:, None, :] - Y[None, :, :]) ** 2).sum(-1))
        K = np.exp(-dists / self.length_scale)

        if eval_gradient:
            if Y is X:
                # Derivative w.r.t. length_scale
                K_gradient = (dists / (self.length_scale ** 2)) * K
                return K, K_gradient[:, :, np.newaxis]
            else:
                return K, np.empty((X.shape[0], Y.shape[0], 0))
        else:
            return K

    def diag(self, X):
        return np.ones(X.shape[0])

    def is_stationary(self):
        return True


class KernelFactory:
    """
    Factory for constructing sklearn-compatible Gaussian Process kernels.
    Supports Gaussian (RBF), Laplacian, and Matern kernels.
    """

    @staticmethod
    def create(kernel_name: str):
        """
        Return the kernel object based on the kernel name.

        Args:
            kernel_name (str): One of ['gaussian', 'laplacian', 'matern52']

        Returns:
            A kernel object compatible with sklearn.gaussian_process.GaussianProcessRegressor
        """
        kernel_name = kernel_name.lower()

        if kernel_name == "gaussian":
            return C(1.0, (1e-3, 1e3)) * RBF(length_scale=1.0, length_scale_bounds=(10, 15))

        elif kernel_name == "laplacian":
            return C(1.0, (1e-3, 1e3)) * LaplacianKernel(length_scale=1.0, length_scale_bounds=(50, 1e2))

        elif kernel_name == "matern52":

            return C(1.0, (1e-3, 1e3)) * Matern(length_scale=1.0, length_scale_bounds=(1e-3, 1e3), nu=2.5)

        else:
            raise ValueError(f"Unsupported kernel: {kernel_name}. Choose from ['gaussian', 'laplacian', 'matern52'].")



"""Download manager for model artifacts.

Handles downloading model weights and other artifacts from various sources:
- Direct HTTP/HTTPS URLs
- Google Drive URLs (via gdown)
- Hugging Face Hub (single files and full snapshots)

Supports setting environment variables to control cache directories.
"""

import json
import shutil
import sys
import tarfile
import urllib.request
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Optional


def _looks_like_html_file(path: Path) -> bool:
    """Detect common HTML placeholders/interstitial pages."""
    if not path.exists() or path.is_dir() or path.stat().st_size == 0:
        return False
    try:
        sample = path.read_bytes()[:512].lstrip().lower()
    except Exception:
        return False
    return sample.startswith(b"<!doctype html") or sample.startswith(b"<html")


def _is_expected_destination_ready(path: Path) -> bool:
    """Return True when a destination artifact already looks usable."""
    if not path.exists():
        return False
    if path.is_dir():
        try:
            return any(path.iterdir())
        except Exception:
            return False
    return not _looks_like_html_file(path)


def _is_google_drive_url(url: str) -> bool:
    value = url.strip().lower()
    return "drive.google.com" in value or "drive.usercontent.google.com" in value


def _normalize_patterns(value: Any) -> Optional[List[str]]:
    if value is None:
        return None
    if isinstance(value, list):
        return [str(item) for item in value if str(item).strip()]
    if isinstance(value, str) and value.strip():
        return [value.strip()]
    return None


def download_google_drive(url: str, dest_path: Path) -> None:
    """Download a file from Google Drive using gdown."""
    try:
        import gdown
    except ImportError as exc:
        raise ImportError(
            "gdown is required for Google Drive downloads. "
            "Install it with: pip install gdown"
        ) from exc

    print(f"Downloading Google Drive file: {url}")
    print(f"     -> {dest_path}")
    dest_path.parent.mkdir(parents=True, exist_ok=True)
    output = gdown.download(url=url, output=str(dest_path), quiet=False, fuzzy=True)
    if not output or not dest_path.exists():
        raise RuntimeError(f"gdown failed to create expected file: {dest_path}")

    if _looks_like_html_file(dest_path):
        raise RuntimeError(
            "Google Drive download returned HTML instead of binary artifact bytes. "
            f"URL: {url}"
        )

    file_size = dest_path.stat().st_size
    print(f"Downloaded {dest_path.name} ({file_size:,} bytes)")


def download_direct(url: str, dest_path: Path) -> None:
    """Download a file directly from a URL."""
    if _is_google_drive_url(url):
        download_google_drive(url, dest_path)
        return

    print(f"Downloading {url}")
    print(f"     -> {dest_path}")

    dest_path.parent.mkdir(parents=True, exist_ok=True)

    try:
        urllib.request.urlretrieve(url, dest_path)
        if _looks_like_html_file(dest_path):
            raise RuntimeError(
                "Downloaded payload appears to be HTML instead of artifact bytes: "
                f"{dest_path}"
            )
        file_size = dest_path.stat().st_size
        print(f"Downloaded {dest_path.name} ({file_size:,} bytes)")
    except Exception as exc:
        raise RuntimeError(f"Failed to download {url}: {exc}") from exc


def download_huggingface(repo_id: str, filename: str, dest_path: Path, cache_dir: str = None) -> None:
    """Download a single file from Hugging Face Hub."""
    try:
        from huggingface_hub import hf_hub_download
    except ImportError as exc:
        raise ImportError(
            "huggingface_hub is not installed. "
            "Install it with: pip install huggingface-hub"
        ) from exc

    print(f"Downloading from Hugging Face: {repo_id}/{filename}")
    print(f"     -> {dest_path}")

    dest_path.parent.mkdir(parents=True, exist_ok=True)

    try:
        hf_path = hf_hub_download(repo_id=repo_id, filename=filename, cache_dir=cache_dir)
        shutil.copy2(hf_path, dest_path)
        if _looks_like_html_file(dest_path):
            raise RuntimeError(
                "Downloaded Hugging Face artifact appears to be HTML instead of bytes: "
                f"{dest_path}"
            )
        file_size = dest_path.stat().st_size
        print(f"Downloaded {dest_path.name} ({file_size:,} bytes)")
    except Exception as exc:
        raise RuntimeError(f"Failed to download from Hugging Face: {exc}") from exc


def download_huggingface_snapshot(
    repo_id: str,
    dest_dir: Path,
    cache_dir: str = None,
    revision: str = None,
    allow_patterns: Optional[List[str]] = None,
    ignore_patterns: Optional[List[str]] = None,
) -> None:
    """Download an entire Hugging Face repo snapshot into a local directory."""
    try:
        from huggingface_hub import snapshot_download
    except ImportError as exc:
        raise ImportError(
            "huggingface_hub is not installed. "
            "Install it with: pip install huggingface-hub"
        ) from exc

    print(f"Downloading Hugging Face snapshot: {repo_id}")
    print(f"     -> {dest_dir}")

    dest_dir.mkdir(parents=True, exist_ok=True)

    kwargs: Dict[str, Any] = {
        "repo_id": repo_id,
        "cache_dir": cache_dir,
        "local_dir": str(dest_dir),
        "local_dir_use_symlinks": False,
    }
    if revision:
        kwargs["revision"] = revision
    if allow_patterns:
        kwargs["allow_patterns"] = allow_patterns
    if ignore_patterns:
        kwargs["ignore_patterns"] = ignore_patterns

    try:
        snapshot_download(**kwargs)
        if not any(dest_dir.iterdir()):
            raise RuntimeError(f"Snapshot download completed but produced an empty directory: {dest_dir}")
        print(f"Downloaded snapshot to {dest_dir}")
    except Exception as exc:
        raise RuntimeError(f"Failed to download Hugging Face snapshot {repo_id}: {exc}") from exc


def load_download_config(config_path: Path) -> Dict[str, Any]:
    """Load download configuration from JSON."""
    with config_path.open("r", encoding="utf-8") as handle:
        return json.load(handle)


def normalize_file_entries(is_hf: int, files: List[Any]) -> List[Dict[str, Any]]:
    """Normalize and validate structured file entries.

    Required schema for each item:
    - {"path": "...", "filename": "..."}

    For direct downloads (isHF == 0), path is a URL.
    For HF downloads (isHF != 0):
    - file mode: path = "repo_id:filename"
    - snapshot mode: path = "repo_id" and set snapshot: 1 (or omit ':' in path)
    """
    del is_hf
    entries: List[Dict[str, Any]] = []

    for item in files:
        if isinstance(item, dict):
            source = item.get("path")
            filename = item.get("filename")
            if not source or not filename:
                raise ValueError(f"File entries must include 'path' and 'filename', got: {item}")

            entry: Dict[str, Any] = {
                "source": str(source),
                "filename": str(filename),
            }

            if "snapshot" in item:
                entry["snapshot"] = bool(item.get("snapshot"))
            if item.get("revision"):
                entry["revision"] = str(item.get("revision"))

            allow_patterns = _normalize_patterns(item.get("allow_patterns"))
            if allow_patterns:
                entry["allow_patterns"] = allow_patterns

            ignore_patterns = _normalize_patterns(item.get("ignore_patterns"))
            if ignore_patterns:
                entry["ignore_patterns"] = ignore_patterns

            entries.append(entry)
            continue

        raise ValueError(
            "Legacy string entries are no longer supported in 'files'. "
            "Use objects like {'path': '...', 'filename': '...'}"
        )

    return entries


def expected_destination_paths(is_hf: int, destination_base: Path, files: List[Any]) -> List[Path]:
    """Compute expected destination file paths from model_cache config."""
    entries = normalize_file_entries(is_hf, files)
    return [destination_base / entry["filename"] for entry in entries]


def extract_archives(destination_base: Path, archive_paths: List[Path], archive_format: str) -> None:
    """Extract archives into destination_base and keep idempotent markers."""
    fmt = (archive_format or "").strip().lower()
    if not fmt:
        return

    for archive_path in archive_paths:
        if not archive_path.exists():
            raise RuntimeError(f"Archive not found for extraction: {archive_path}")

        marker = destination_base / f".extract_complete_{archive_path.name}"
        if marker.exists():
            print(f"Extraction already completed for {archive_path.name}, skipping.")
            continue

        print(f"Extracting {archive_path.name} ({archive_format}) into {destination_base}")
        destination_base.mkdir(parents=True, exist_ok=True)

        if fmt == "zip":
            with zipfile.ZipFile(archive_path, "r") as zip_ref:
                zip_ref.extractall(destination_base)
        elif fmt in {"tar", "tar.gz", "tgz"}:
            mode = "r:gz" if fmt in {"tar.gz", "tgz"} else "r"
            with tarfile.open(archive_path, mode) as tar_ref:
                tar_ref.extractall(destination_base)
        else:
            raise ValueError(f"Unsupported archive format '{archive_format}' in model_cache.json")

        marker.write_text("ok\n", encoding="utf-8")
        print(f"Extracted {archive_path.name}")


def download_model_artifacts(
    config_path: Path,
    workspace_root: Path,
    env_vars: Dict[str, str] = None,
    skip_if_complete: bool = True,
) -> None:
    """Download all model artifacts specified in config."""
    import os

    config = load_download_config(config_path)
    is_hf = config.get("isHF", 0)
    archive_format = config.get("format", "")
    destination = config.get("destination", "")
    files = config.get("files", [])

    if not destination or not files:
        print("No files to download (empty config).")
        return

    if env_vars:
        for key, value in env_vars.items():
            if value.startswith("./") or value.startswith(".\\"):
                value = str(workspace_root / value[2:])
            os.environ[key] = value
            print(f"Set {key}={value}")

    dest_base = workspace_root / destination
    entries = normalize_file_entries(is_hf, files)
    expected_paths = [dest_base / entry["filename"] for entry in entries]
    all_downloads_present = bool(expected_paths) and all(
        _is_expected_destination_ready(path) for path in expected_paths
    )

    if skip_if_complete and all_downloads_present:
        print(f"All expected artifacts already exist in {dest_base}. Skipping download phase.")

    if not all_downloads_present:
        if is_hf == 0:
            for entry in entries:
                url = entry["source"]
                dest_path = dest_base / entry["filename"]

                if dest_path.exists():
                    if _looks_like_html_file(dest_path):
                        print(f"{dest_path} is an HTML placeholder. Re-downloading.")
                        dest_path.unlink()
                    else:
                        print(f"{dest_path} already exists, skipping.")
                        continue

                download_direct(url, dest_path)
        else:
            for entry in entries:
                file_spec = entry["source"]
                dest_path = dest_base / entry["filename"]
                hf_cache = os.environ.get("HF_HOME")

                snapshot_mode = bool(entry.get("snapshot")) or (":" not in file_spec)

                if snapshot_mode:
                    if ":" in file_spec:
                        raise ValueError(
                            "HF snapshot entries must use a repo_id without ':filename'. "
                            f"Got: {file_spec}"
                        )

                    if _is_expected_destination_ready(dest_path):
                        print(f"{dest_path} already exists, skipping snapshot download.")
                        continue

                    download_huggingface_snapshot(
                        repo_id=file_spec,
                        dest_dir=dest_path,
                        cache_dir=hf_cache,
                        revision=entry.get("revision"),
                        allow_patterns=entry.get("allow_patterns"),
                        ignore_patterns=entry.get("ignore_patterns"),
                    )
                    continue

                if ":" not in file_spec:
                    raise ValueError(f"HF file spec must be 'repo_id:filename', got: {file_spec}")

                repo_id, source_filename = file_spec.split(":", 1)

                if dest_path.exists():
                    if _looks_like_html_file(dest_path):
                        print(f"{dest_path} is an HTML placeholder. Re-downloading.")
                        dest_path.unlink()
                    else:
                        print(f"{dest_path} already exists, skipping.")
                        continue

                download_huggingface(repo_id, source_filename, dest_path, cache_dir=hf_cache)

    if archive_format:
        extract_archives(dest_base, expected_paths, archive_format)


if __name__ == "__main__":
    if len(sys.argv) > 1:
        config_file = Path(sys.argv[1])
        workspace = Path.cwd()
        download_model_artifacts(config_file, workspace)
    else:
        print("Usage: python downloads_manager.py <config_path>")



from pathlib import Path

import numpy as np
import pandas as pd
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.metrics import mean_squared_error, r2_score
from sklearn.preprocessing import StandardScaler

try:
    from .kernel import KernelFactory
except ImportError:
    from kernel import KernelFactory


def _read_targets_dataframe(target_path: str) -> pd.DataFrame:
    try:
        return pd.read_csv(target_path)
    except Exception:
        return pd.read_csv(target_path, sep=",", engine="python")


def _resolve_target_column(targets_df: pd.DataFrame) -> str:
    if len(targets_df.columns) < 2:
        raise ValueError(
            "Target file must contain at least two columns: molecules and target values."
        )
    return str(targets_df.columns[1])


def evaluate(
    embeddings_path: str,
    target_path: str,
    splits_dir: str,
    n_splits: int = 5,
    compute_rmse: bool = True,
    kernel_name: str = "gaussian",
    results_dir: str | None = None,
):
    """Evaluate Gaussian Process regression across pre-saved splits."""
    X = np.load(embeddings_path, allow_pickle=True)

    targets_df = _read_targets_dataframe(target_path)
    target_column = _resolve_target_column(targets_df)
    y = targets_df[target_column].to_numpy()
    print(f"Loaded {len(y)} samples from target column: {target_column}.")

    kernel = KernelFactory.create(kernel_name)
    model = GaussianProcessRegressor(
        kernel=kernel,
        n_restarts_optimizer=10,
        normalize_y=True,
        alpha=1e-6,
    )

    results = []
    train_r2, test_r2 = [], []
    if compute_rmse:
        train_rmse, test_rmse = [], []

    train_dir = Path(splits_dir) / "train"
    test_dir = Path(splits_dir) / "test"

    for i in range(1, n_splits + 1):
        train_idx = np.load(train_dir / f"train_split_{i}.npy")
        test_idx = np.load(test_dir / f"test_split_{i}.npy")

        X_train, X_test = X[train_idx], X[test_idx]
        y_train, y_test = y[train_idx], y[test_idx]

        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)
        X_test = scaler.transform(X_test)

        print(f"\nSplit {i}: Train={len(train_idx)}, Test={len(test_idx)}")

        model.fit(X_train, y_train)
        params = model.kernel_.get_params()

        y_pred_train = model.predict(X_train)
        y_pred_test = model.predict(X_test)

        r2_train = r2_score(y_train, y_pred_train)
        r2_test = r2_score(y_test, y_pred_test)
        train_r2.append(r2_train)
        test_r2.append(r2_test)

        row = {
            "split": i,
            "train_size": len(train_idx),
            "test_size": len(test_idx),
            "r2_train": r2_train,
            "r2_test": r2_test,
        }

        for key, value in params.items():
            row[key] = value

        if compute_rmse:
            rmse_train = np.sqrt(mean_squared_error(y_train, y_pred_train))
            rmse_test = np.sqrt(mean_squared_error(y_test, y_pred_test))
            train_rmse.append(rmse_train)
            test_rmse.append(rmse_test)
            row.update({"rmse_train": rmse_train, "rmse_test": rmse_test})

        results.append(row)

        print(f"R2 Train={r2_train:.3f}, R2 Test={r2_test:.3f}")
        if compute_rmse:
            print(f"RMSE Train={rmse_train:.3f}, RMSE Test={rmse_test:.3f}")

    df_results = pd.DataFrame(results)

    summary = {
        "split": "mean +/- std",
        "train_size": "-",
        "test_size": "-",
        "r2_train": f"{df_results['r2_train'].mean():.3f} +/- {df_results['r2_train'].std():.3f}",
        "r2_test": f"{df_results['r2_test'].mean():.3f} +/- {df_results['r2_test'].std():.3f}",
    }
    if compute_rmse:
        summary.update(
            {
                "rmse_train": f"{df_results['rmse_train'].mean():.3f} +/- {df_results['rmse_train'].std():.3f}",
                "rmse_test": f"{df_results['rmse_test'].mean():.3f} +/- {df_results['rmse_test'].std():.3f}",
            }
        )

    df_results = pd.concat([df_results, pd.DataFrame([summary])], ignore_index=True)

    embed_name = Path(embeddings_path).name
    model_id = embed_name.replace(".npy", "")
    if model_id:
        model_id = model_id[0].upper() + model_id[1:]
    else:
        model_id = "Unknown"

    if results_dir:
        output_dir = Path(results_dir)
    else:
        output_dir = Path("Results") / model_id
    output_dir.mkdir(parents=True, exist_ok=True)

    output_csv = output_dir / f"results_{kernel_name}.csv"
    df_results.to_csv(output_csv, index=False)

    print(f"\nResults saved to {output_csv.resolve()}")
    print(f"Kernel used: {kernel_name}\n")

    return df_results



import json
import re
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple


def venv_python_path(venv_dir: Path) -> Path:
    if (venv_dir / "Scripts" / "python.exe").exists():
        return venv_dir / "Scripts" / "python.exe"
    return venv_dir / "bin" / "python"


def _sanitize_token(value: str) -> str:
    normalized = re.sub(r"[^A-Za-z0-9._-]+", "_", value).strip("._-")
    return normalized or "input"


def create_embeddings_run_dir(workspace_root: Path, model_name: str, input_file: Path) -> Tuple[Path, str]:
    base_dir = workspace_root / "data" / "runs" / "embeddings" / model_name.upper()
    base_dir.mkdir(parents=True, exist_ok=True)

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    token = _sanitize_token(input_file.stem)
    run_id = f"{timestamp}_{token}"
    run_dir = base_dir / run_id

    suffix = 2
    while run_dir.exists():
        run_id = f"{timestamp}_{token}_{suffix}"
        run_dir = base_dir / run_id
        suffix += 1

    run_dir.mkdir(parents=True, exist_ok=False)
    return run_dir, run_id


def _load_model_metadata(metadata_root: Path, model_name: str) -> Dict[str, Any]:
    metadata_dir = metadata_root / model_name
    candidates = [metadata_dir / "model.json", metadata_dir / f"{model_name}.json"]
    for path in candidates:
        if path.exists():
            with path.open("r", encoding="utf-8") as handle:
                return json.load(handle)
    raise FileNotFoundError(f"Metadata not found for model {model_name} in {metadata_dir}")


def build_embeddings_command(
    workspace_root: Path,
    model_name: str,
    input_file: Path,
    input_format: str,
    output_file: Optional[Path] = None,
    extra_args: Optional[List[str]] = None,
    metadata_root: Optional[Path] = None,
) -> Tuple[str, List[str], str, Path, Path, str]:
    """Build a venv-scoped embeddings command for a model.

    Returns (program, args, cwd, resolved_output_path, run_dir, run_id).
    """
    model_dir = workspace_root / "models" / model_name
    embeddings_script = model_dir / "embeddings.py"
    venv_dir = model_dir / ".venv"
    python_exe = venv_python_path(venv_dir)

    if not embeddings_script.exists():
        raise FileNotFoundError(f"Embeddings script not found: {embeddings_script}")
    if not python_exe.exists():
        raise FileNotFoundError(f"Model venv python not found: {python_exe}")
    if not input_file.exists():
        raise FileNotFoundError(f"Input file not found: {input_file}")

    if output_file is None:
        run_dir, run_id = create_embeddings_run_dir(workspace_root, model_name, input_file)
        resolved_output = run_dir / f"{input_file.stem}.npy"
    else:
        resolved_output = output_file
        run_dir = resolved_output.parent
        run_id = run_dir.name
        run_dir.mkdir(parents=True, exist_ok=True)

    resolved_output.parent.mkdir(parents=True, exist_ok=True)

    resolved_metadata_root = metadata_root or (workspace_root / "metadata")
    metadata = _load_model_metadata(resolved_metadata_root, model_name)
    cli_version = int(metadata.get("embeddingsCliVersion", 1))
    if model_name.upper() == "CLAMP" and cli_version < 2:
        raise RuntimeError(
            "CLAMP requires embeddingsCliVersion >= 2 and the new embeddings CLI protocol."
        )

    if cli_version >= 2:
        model_input_type = str(metadata.get("inputType", "smiles")).strip().lower()
        normalized_input_format = str(input_format).strip().lower()
        args = [
            str(embeddings_script),
            "--file",
            str(input_file),
            "--input-format",
            normalized_input_format,
            "--model-input-type",
            model_input_type,
            "--output-dir",
            str(resolved_output.parent),
            "--output-stem",
            resolved_output.stem,
        ]
    else:
        # Legacy embeddings CLI compatibility.
        args = [
            str(embeddings_script),
            "--input",
            str(input_file),
            "--output",
            str(resolved_output),
        ]

    if extra_args:
        args.extend(extra_args)

    return str(python_exe), args, str(model_dir), resolved_output, run_dir, run_id



from abc import ABC, abstractmethod

import torch
from torch import nn
from torch.nn.modules.loss import CrossEntropyLoss, MSELoss
from transformers.models.bert.modeling_bert import BertLMPredictionHead

from molbert.tasks.heads import PhysChemHead, FinetuneHead, IsSameHead


class BaseTask(nn.Module, ABC):
    def __init__(self, name):
        super().__init__()
        self.name = name

    @abstractmethod
    def forward(self, sequence_output, pooled_output):
        raise NotImplementedError

    @abstractmethod
    def compute_loss(self, batch_labels, batch_predictions) -> torch.Tensor:
        raise NotImplementedError


class FinetuneTask(BaseTask):
    def __init__(self, name, config):
        super().__init__(name)
        self.loss: nn.Module
        self.mode = config.mode
        if config.mode == 'regression':
            self.loss = MSELoss()
        elif config.mode == 'classification':
            self.loss = CrossEntropyLoss()
        else:
            raise ValueError(f'config.mode must be in [regression, classification] but was {config.mode}')

        self.head = FinetuneHead(config)

    def forward(
        self,
        sequence_output,
        pooled_output,
    ):
        return self.head(pooled_output)

    def compute_loss(self, batch_labels, batch_predictions) -> torch.Tensor:
        predictions = batch_predictions[self.name]
        labels = batch_labels[self.name]

        if self.mode == 'classification':
            labels = labels.long().squeeze(1)

        return self.loss(predictions, labels)


class PhyschemTask(BaseTask):
    def __init__(self, name, config):
        super().__init__(name)
        self.loss = MSELoss()

        self.physchem_head = PhysChemHead(config)

    def forward(self, sequence_output, pooled_output):
        return self.physchem_head(pooled_output)

    def compute_loss(self, batch_labels, batch_predictions) -> torch.Tensor:
        return self.loss(batch_predictions[self.name], batch_labels[self.name])


class MaskedLMTask(BaseTask):
    def __init__(self, name, config):
        super().__init__(name)
        self.loss = CrossEntropyLoss(ignore_index=-1)
        self.vocab_size = config.vocab_size
        self.masked_lm_head = BertLMPredictionHead(config)

    def forward(self, sequence_output, pooled_output):
        return self.masked_lm_head(sequence_output)

    def compute_loss(self, batch_labels, batch_predictions) -> torch.Tensor:
        return self.loss(
            batch_predictions['masked_lm'].view(-1, self.vocab_size), batch_labels['lm_label_ids'].view(-1)
        )


class IsSameTask(BaseTask):
    def __init__(self, name, config):
        super().__init__(name)
        self.loss = CrossEntropyLoss(ignore_index=-1)
        self.is_same_head = IsSameHead(config)

    def forward(self, sequence_output, pooled_output):
        return self.is_same_head(pooled_output)

    def compute_loss(self, batch_labels, batch_predictions) -> torch.Tensor:
        return self.loss(batch_predictions[self.name].view(-1, 2), batch_labels[self.name].view(-1))



"""Workspace launcher for GP regression benchmark runs.

This script is called by the GUI and handles:
1. Provisioning/reusing a dedicated benchmark virtual environment.
2. Installing benchmark dependencies from requirements.txt.
3. Executing the GP regression pipeline (single or batch mode).
"""

from __future__ import annotations

import argparse
import hashlib
import json
import os
import re
import subprocess
import sys
from datetime import datetime, timezone
from pathlib import Path


BENCHMARK_NAME = "GP_regressor"
DEPS_STATE_FILE = ".deps_state.json"
DATA_SENTINEL = ".morehub_data.json"
RUN_META_FILE = ".run_meta.json"


def _log(message: str) -> None:
    print(message, flush=True)


def _cmd_display(cmd: list[str]) -> str:
    if os.name == "nt":
        return subprocess.list2cmdline(cmd)
    return " ".join(cmd)


def _run_checked(cmd: list[str], cwd: Path | None = None) -> None:
    _log(f"Running: {_cmd_display(cmd)}")
    subprocess.run(cmd, cwd=str(cwd) if cwd else None, check=True)


def _sha256(path: Path) -> str:
    digest = hashlib.sha256()
    with path.open("rb") as handle:
        for chunk in iter(lambda: handle.read(1024 * 1024), b""):
            digest.update(chunk)
    return digest.hexdigest()


def _venv_python(venv_dir: Path) -> Path:
    if os.name == "nt":
        return venv_dir / "Scripts" / "python.exe"
    return venv_dir / "bin" / "python"


def _ensure_venv(venv_dir: Path) -> Path:
    python_path = _venv_python(venv_dir)
    if python_path.exists():
        _log(f"Benchmark virtual environment already exists: {venv_dir}")
        return python_path

    _log(f"Creating benchmark virtual environment: {venv_dir}")
    _run_checked([sys.executable, "-m", "venv", str(venv_dir)])

    if not python_path.exists():
        raise RuntimeError(f"Failed to create benchmark Python at {python_path}")
    return python_path


def _load_json(path: Path) -> dict:
    if not path.exists():
        return {}
    try:
        return json.loads(path.read_text(encoding="utf-8"))
    except Exception:
        return {}


def _save_json(path: Path, payload: dict) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(payload, indent=2) + "\n", encoding="utf-8")


def _ensure_dependencies(venv_python: Path, requirements_path: Path, state_path: Path) -> None:
    requirements_hash = _sha256(requirements_path)
    state = _load_json(state_path)

    if (
        state.get("requirements_sha256") == requirements_hash
        and state.get("requirements_path") == str(requirements_path)
        and state.get("venv_python") == str(venv_python)
    ):
        _log("Benchmark dependencies already satisfied. Skipping pip install.")
        return

    _log("Installing benchmark dependencies...")
    _run_checked([str(venv_python), "-m", "pip", "install", "--upgrade", "pip"])
    _run_checked([str(venv_python), "-m", "pip", "install", "-r", str(requirements_path)])

    _save_json(
        state_path,
        {
            "benchmark": BENCHMARK_NAME,
            "requirements_path": str(requirements_path),
            "requirements_sha256": requirements_hash,
            "venv_python": str(venv_python),
            "installed_at": datetime.now(timezone.utc).isoformat(),
        },
    )


def _safe_name(value: str) -> str:
    return re.sub(r"[^A-Za-z0-9_.-]+", "_", value.strip()).strip("_") or "run"


def _next_run_root(benchmark_root: Path) -> tuple[Path, int]:
    max_index = 0
    pattern = re.compile(r"^run_(\d+)$")

    for candidate in benchmark_root.iterdir():
        if not candidate.is_dir():
            continue
        match = pattern.match(candidate.name)
        if not match:
            continue
        max_index = max(max_index, int(match.group(1)))

    run_index = max_index + 1
    run_root = benchmark_root / f"run_{run_index}"
    run_root.mkdir(parents=True, exist_ok=False)
    return run_root, run_index


def _write_run_meta(run_root: Path, payload: dict) -> Path:
    path = run_root / RUN_META_FILE
    _save_json(path, payload)
    return path


def _derive_model_label(embeddings_path: Path, embeddings_root: Path) -> str:
    try:
        rel = embeddings_path.resolve().relative_to(embeddings_root.resolve())
        if rel.parts:
            return _safe_name(rel.parts[0])
    except ValueError:
        pass
    return _safe_name(embeddings_path.parent.name or "model")


def _build_gp_command(
    venv_python: Path,
    gp_main_script: Path,
    embeddings_path: Path,
    target_path: Path,
    splits_dir: Path,
    kernels: list[str],
    compute_rmse: bool,
    n_splits: int,
    train_ratio: float,
    regenerate_splits: bool,
    results_dir: Path,
) -> list[str]:
    cmd = [
        str(venv_python),
        str(gp_main_script),
        "--embeddings_path",
        str(embeddings_path),
        "--target_path",
        str(target_path),
        "--splits_dir",
        str(splits_dir),
        "--kernel_names",
        *kernels,
        "--compute_rmse",
        "1" if compute_rmse else "0",
        "--n_splits",
        str(n_splits),
        "--train_ratio",
        f"{train_ratio}",
        "--results_dir",
        str(results_dir),
    ]
    if regenerate_splits:
        cmd.append("--regenerate_splits")
    return cmd


def _write_manifest(path: Path, payload: dict) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(payload, indent=2) + "\n", encoding="utf-8")


def _new_case_dir(run_root: Path, model_label: str) -> Path:
    token = _safe_name(model_label)
    case_dir = run_root / token
    suffix = 2
    while case_dir.exists():
        case_dir = run_root / f"{token}_{suffix}"
        suffix += 1
    case_dir.mkdir(parents=True, exist_ok=False)
    return case_dir


def _run_single_case(
    *,
    venv_python: Path,
    gp_main_script: Path,
    gp_source_dir: Path,
    run_root: Path,
    data_root: Path,
    embeddings_path: Path,
    target_path: Path,
    kernels: list[str],
    compute_rmse: bool,
    n_splits: int,
    train_ratio: float,
    regenerate_splits: bool,
    mode: str,
    model_label: str,
    case_label: str,
) -> dict:
    case_dir = _new_case_dir(run_root, model_label)
    splits_dir = case_dir / "splits"
    results_dir = case_dir / "results"

    command = _build_gp_command(
        venv_python=venv_python,
        gp_main_script=gp_main_script,
        embeddings_path=embeddings_path,
        target_path=target_path,
        splits_dir=splits_dir,
        kernels=kernels,
        compute_rmse=compute_rmse,
        n_splits=n_splits,
        train_ratio=train_ratio,
        regenerate_splits=regenerate_splits,
        results_dir=results_dir,
    )

    exit_code = 0
    try:
        _run_checked(command, cwd=gp_source_dir)
        _log(f"[GP] Run completed successfully: {case_dir}")
    except subprocess.CalledProcessError as exc:
        exit_code = int(exc.returncode)
        _log(f"[GP] Run failed with exit code {exit_code}: {case_dir}")

    manifest = {
        "tool": "gp_regression",
        "mode": mode,
        "created_at": datetime.now(timezone.utc).isoformat(),
        "data_root": str(data_root),
        "run_root": str(run_root),
        "model_label": model_label,
        "case_label": case_label,
        "input": {
            "embeddings_path": str(embeddings_path),
            "target_path": str(target_path),
        },
        "config": {
            "kernels": kernels,
            "compute_rmse": compute_rmse,
            "n_splits": n_splits,
            "train_ratio": train_ratio,
            "regenerate_splits": regenerate_splits,
        },
        "command": command,
        "output": {
            "case_dir": str(case_dir),
            "splits_dir": str(splits_dir),
            "results_dir": str(results_dir),
        },
        "status": "success" if exit_code == 0 else "failed",
        "exit_code": exit_code,
    }
    _write_manifest(case_dir / "manifest.json", manifest)
    _log(f"[GP] Run manifest: {case_dir / 'manifest.json'}")

    return {
        "status": manifest["status"],
        "exit_code": exit_code,
        "manifest_path": str(case_dir / "manifest.json"),
        "case_dir": str(case_dir),
        "model_label": model_label,
        "case_label": case_label,
    }


def _run_single_mode(
    args: argparse.Namespace,
    venv_python: Path,
    gp_main_script: Path,
    gp_source_dir: Path,
    run_root: Path,
    data_root: Path,
    embeddings_root: Path,
) -> dict:
    embeddings_path = Path(args.embeddings_path).resolve()
    target_path = Path(args.targets_path).resolve()
    if not embeddings_path.exists():
        raise FileNotFoundError(f"Embeddings file does not exist: {embeddings_path}")
    if not target_path.exists():
        raise FileNotFoundError(f"Targets file does not exist: {target_path}")

    try:
        embeddings_path.relative_to(embeddings_root.resolve())
    except ValueError:
        raise RuntimeError(
            "Single embeddings file must be inside <data_instance>/data/embeddings. "
            f"Provided: {embeddings_path}"
        )

    model_label = _derive_model_label(embeddings_path, embeddings_root)
    item = _run_single_case(
        venv_python=venv_python,
        gp_main_script=gp_main_script,
        gp_source_dir=gp_source_dir,
        run_root=run_root,
        data_root=data_root,
        embeddings_path=embeddings_path,
        target_path=target_path,
        kernels=args.kernel,
        compute_rmse=args.compute_rmse == "1",
        n_splits=args.n_splits,
        train_ratio=args.train_ratio,
        regenerate_splits=args.regenerate_splits == "1",
        mode="single",
        model_label=model_label,
        case_label=embeddings_path.stem,
    )

    return {
        "exit_code": item["exit_code"],
        "items": [item],
        "failures": 0 if item["exit_code"] == 0 else 1,
        "total": 1,
    }


def _run_batch_mode(
    args: argparse.Namespace,
    venv_python: Path,
    gp_main_script: Path,
    gp_source_dir: Path,
    run_root: Path,
    data_root: Path,
    embeddings_root: Path,
) -> dict:
    root_path = Path(args.root_path).resolve()
    target_path = Path(args.targets_path).resolve()

    if not root_path.exists() or not root_path.is_dir():
        raise FileNotFoundError(f"Batch root directory does not exist: {root_path}")
    if not target_path.exists():
        raise FileNotFoundError(f"Targets file does not exist: {target_path}")

    try:
        root_path.relative_to(embeddings_root.resolve())
    except ValueError:
        raise RuntimeError(
            "Batch root must be inside <data_instance>/data/embeddings. "
            f"Provided: {root_path}"
        )

    embeddings_files = sorted(p for p in root_path.rglob("*.npy") if p.is_file())
    if not embeddings_files:
        raise FileNotFoundError(f"No .npy embeddings files found under: {root_path}")

    _log(f"[GP] Batch mode: found {len(embeddings_files)} embeddings files")

    failures = 0
    items = []
    for embedding_path in embeddings_files:
        model_label = _derive_model_label(embedding_path, embeddings_root)
        case_label = embedding_path.stem

        item = _run_single_case(
            venv_python=venv_python,
            gp_main_script=gp_main_script,
            gp_source_dir=gp_source_dir,
            run_root=run_root,
            data_root=data_root,
            embeddings_path=embedding_path,
            target_path=target_path,
            kernels=args.kernel,
            compute_rmse=args.compute_rmse == "1",
            n_splits=args.n_splits,
            train_ratio=args.train_ratio,
            regenerate_splits=args.regenerate_splits == "1",
            mode="batch",
            model_label=model_label,
            case_label=case_label,
        )
        items.append(item)
        if item["exit_code"] != 0:
            failures += 1

    session_manifest = {
        "tool": "gp_regression",
        "mode": "batch",
        "created_at": datetime.now(timezone.utc).isoformat(),
        "data_root": str(data_root),
        "run_root": str(run_root),
        "batch_root": str(root_path),
        "target_path": str(target_path),
        "summary": {
            "total": len(items),
            "failures": failures,
            "successes": len(items) - failures,
        },
        "items": items,
    }
    _write_manifest(run_root / "manifest.json", session_manifest)
    _log(f"[GP] Session manifest: {run_root / 'manifest.json'}")
    _log(f"[GP] Batch summary: failures={failures}, total={len(items)}")

    return {
        "exit_code": 1 if failures > 0 else 0,
        "items": items,
        "failures": failures,
        "total": len(items),
    }


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Run GP benchmark with managed benchmark virtual environment")
    parser.add_argument("--data_root", required=True, help="Absolute data instance root path (data_<name>)")
    parser.add_argument("--mode", choices=["single", "batch"], required=True)

    parser.add_argument("--embeddings_path", default=None, help="Single mode: embeddings .npy path")
    parser.add_argument("--targets_path", default=None, help="Targets file path")
    parser.add_argument("--root_path", default=None, help="Batch mode: root folder to scan for .npy files")

    parser.add_argument("--kernel", action="append", choices=["gaussian", "laplacian", "matern52"], default=[])
    parser.add_argument("--compute_rmse", choices=["0", "1"], default="1")
    parser.add_argument("--regenerate_splits", choices=["0", "1"], default="0")
    parser.add_argument("--n_splits", type=int, default=5)
    parser.add_argument("--train_ratio", type=float, default=0.7)

    args = parser.parse_args()

    if not args.targets_path:
        parser.error("--targets_path is required")

    if args.mode == "single":
        if not args.embeddings_path:
            parser.error("single mode requires --embeddings_path")
    if args.mode == "batch" and not args.root_path:
        parser.error("batch mode requires --root_path")

    if not args.kernel:
        args.kernel = ["gaussian"]

    if args.n_splits < 1:
        parser.error("--n_splits must be >= 1")
    if not (0.0 < args.train_ratio <= 1.0):
        parser.error("--train_ratio must be in (0.0, 1.0]")

    return args


def main() -> int:
    args = _parse_args()

    data_root = Path(args.data_root).resolve()
    if not data_root.exists() or not data_root.is_dir():
        raise FileNotFoundError(f"Data root directory does not exist: {data_root}")
    if not (data_root / DATA_SENTINEL).exists():
        raise RuntimeError(f"Invalid data instance (missing {DATA_SENTINEL}): {data_root}")

    benchmark_root = data_root / "data" / "benchmarks" / BENCHMARK_NAME
    embeddings_root = data_root / "data" / "embeddings"
    venv_dir = benchmark_root / ".venv"

    gp_source_dir = Path(__file__).resolve().parent
    gp_main_script = gp_source_dir / "main.py"
    requirements_path = gp_source_dir / "requirements.txt"

    benchmark_root.mkdir(parents=True, exist_ok=True)
    embeddings_root.mkdir(parents=True, exist_ok=True)

    if not gp_main_script.exists():
        raise FileNotFoundError(f"GP main script not found: {gp_main_script}")
    if not requirements_path.exists():
        raise FileNotFoundError(f"GP requirements file not found: {requirements_path}")

    run_root, run_index = _next_run_root(benchmark_root)
    created_at = datetime.now(timezone.utc).isoformat()

    run_meta = {
        "format": "morehub-gp-run",
        "version": 1,
        "run_index": run_index,
        "mode": args.mode,
        "created_at": created_at,
        "status": "running",
        "data_root": str(data_root),
        "run_root": str(run_root),
    }
    _write_run_meta(run_root, run_meta)

    _log(f"[GP] Data root: {data_root}")
    _log(f"[GP] Run root: {run_root}")

    venv_python = _ensure_venv(venv_dir)
    _ensure_dependencies(venv_python, requirements_path, benchmark_root / DEPS_STATE_FILE)

    if args.mode == "single":
        outcome = _run_single_mode(
            args,
            venv_python,
            gp_main_script,
            gp_source_dir,
            run_root,
            data_root,
            embeddings_root,
        )
    else:
        outcome = _run_batch_mode(
            args,
            venv_python,
            gp_main_script,
            gp_source_dir,
            run_root,
            data_root,
            embeddings_root,
        )

    finished_at = datetime.now(timezone.utc).isoformat()
    run_meta.update(
        {
            "finished_at": finished_at,
            "status": "success" if outcome["exit_code"] == 0 else "failed",
            "summary": {
                "total": outcome.get("total", 0),
                "failures": outcome.get("failures", 0),
                "successes": max(0, outcome.get("total", 0) - outcome.get("failures", 0)),
            },
            "items": outcome.get("items", []),
        }
    )
    _write_run_meta(run_root, run_meta)

    return int(outcome["exit_code"])


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from pathlib import Path
from typing import Callable

from PySide6.QtCore import QObject, Qt
from PySide6.QtWidgets import QMainWindow

from benchmarks.gp_regressor_widget import GPRegressorWidget


class BenchmarkGP(QObject):
    """Middle-layer interface between app.py and the GP benchmark UI."""

    def __init__(
        self,
        *,
        package_root: Path,
        workspace_provider: Callable[[], Path | None],
        owner: QMainWindow | None = None,
    ) -> None:
        super().__init__(owner)
        self._package_root = package_root
        self._workspace_provider = workspace_provider
        self._widget: GPRegressorWidget | None = None

    def show(self) -> None:
        if self._widget is None:
            self._widget = GPRegressorWidget(
                workspace_provider=self._workspace_provider,
                package_root=self._package_root,
                parent=None,
            )
            self._widget.setWindowFlag(Qt.Window, True)
            self._widget.destroyed.connect(self._clear_widget)

        self._widget.refresh_workspace_state()
        self._widget.show()
        self._widget.raise_()
        self._widget.activateWindow()

    def refresh_workspace_state(self) -> None:
        if self._widget is not None:
            self._widget.refresh_workspace_state()

    def _clear_widget(self, *_args) -> None:
        self._widget = None



import html
import json
import re
import sys
import hashlib
import shutil
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from PySide6.QtCore import QProcess, Qt
from PySide6.QtGui import QFont, QIcon
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QDialog,
    QFileDialog,
    QLabel,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QTextEdit,
)

try:
    from .runner import build_embeddings_command
    from .ui.mainui import Ui_MainWindow
    from .ui.new_data_form import Ui_Form as Ui_NewDataForm
except ImportError:
    from runner import build_embeddings_command
    from ui.mainui import Ui_MainWindow
    from ui.new_data_form import Ui_Form as Ui_NewDataForm

try:
    from .data_manager import (
        create_data_instance,
        find_data_root,
        is_data_instance,
        model_embeddings_dir,
        resolve_data_source,
    )
except ImportError:
    from data_manager import (
        create_data_instance,
        find_data_root,
        is_data_instance,
        model_embeddings_dir,
        resolve_data_source,
    )

try:
    from benchmarks.benchmark_gp import BenchmarkGP
except Exception:
    BenchmarkGP = None


WORKSPACE_MARKER_NAME = ".morehub_workspace.json"
APP_STATE_DIR = Path.home() / ".morehub"
APP_STATE_FILE = APP_STATE_DIR / "state.json"


class App(QMainWindow):
    """Main GUI application that owns UI and button behaviors."""

    def __init__(self) -> None:
        super().__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self._apply_visual_consistency()

        self.package_root = Path(__file__).resolve().parent
        self.metadata_root = self.package_root / "metadata"
        self.workspace_root: Path | None = None

        # State tracking for processes, consoles, and input selections
        self._process_by_model = {}
        self._console_by_model = {}
        self._model_by_button = {}
        self._status_widgets_by_model = {}
        self._input_button_to_model = {}
        self._input_label_by_model = {}
        self._selected_input_by_model = {}
        self._global_data_root: Path | None = None
        self._run_embeddings_after_setup = {}
        self._format_widgets_by_model = {}
        self._embeddings_extra_args_builders = {}
        self._grover_option_widgets = {}
        self._chemberta_variant_widgets = {}
        self._novomolgen_variant_widgets = {}
        self._molgen_variant_widgets = {}
        self._molclr_backbone_widgets = {}
        self._overwrite_widget_by_model = {}
        self._active_embeddings_runs = {}
        self._benchmark_gp = BenchmarkGP(package_root=self.package_root, workspace_provider=lambda: self.workspace_root, owner=self) if BenchmarkGP is not None else None

        self._bind_run_buttons()
        self._bind_input_buttons()
        self._bind_format_controls()
        self._bind_embeddings_option_controls()
        self._bind_menu_actions()
        self._bind_status_widgets()
        self._clear_all_data_selections()
        self.ui.tabWidget.currentChanged.connect(self._on_model_tab_changed)
        self._load_persisted_workspace()
        self._update_window_title()
        self._refresh_all_statuses()
        if self.workspace_root is None:
            self._broadcast_log("[WORKSPACE] Select File -> New.. or File -> Load to choose a working directory.")

    def _apply_visual_consistency(self) -> None:
        """Normalize font and indicator sizing across machines/styles."""
        radio_button_names = [
            "radioButton",
            "radioButton_2",
            "radioButton_3",
            "radioButton_4",
            "radioButton_5",
            "radioButton_6",
            "radioButton_7",
            "radioButton_8",
            "radioButton_9",
            "radioButton_10",
            "radioButton_11",
            "radioButton_12",
            "radioButton_13",
            "radioButton_14",
            "radioButton_15",
            "radioButton_16",
            "radioButton_17",
            "radioButton_18",
            "radioButton_19",
            "radioButton_20",
            "radioButton_21",
            "radioButton_22",
            "radioButton_23",
            "radioButton_25",
            "radioButton_26",
            "radioButton_27",
            "radioButton_28",
            "radioButton_24",
            "radioButton_29",
            "radioButton_30",
            "radioButton_31",
            "radioButton_32",
            "radioButton_33",
            "radioButton_34",
            "radioButton_35",
            "radioButton_36",
            "radioButton_37",
            "radioButton_38",
        ]

        radio_font = QFont()
        radio_font.setFamilies(["Oswald", "Nirmala UI", "Segoe UI", "Arial"])
        radio_font.setPointSize(10)

        for name in radio_button_names:
            widget = getattr(self.ui, name, None)
            if not isinstance(widget, QRadioButton):
                continue
            widget.setFont(radio_font)
            widget.setStyleSheet(
                "QRadioButton { spacing: 8px; } "
                "QRadioButton::indicator { width: 14px; height: 14px; }"
            )
    def _bind_menu_actions(self) -> None:
        if hasattr(self.ui, "actionNew"):
            self.ui.actionNew.triggered.connect(self._on_new_workspace_clicked)
        if hasattr(self.ui, "actionLoad"):
            self.ui.actionLoad.triggered.connect(self._on_load_workspace_clicked)
        if hasattr(self.ui, "actionNew_Data_Setup"):
            self.ui.actionNew_Data_Setup.triggered.connect(self._on_new_data_setup_clicked)

        benchmark_action = getattr(self.ui, "actionGP_Regression", None)
        if benchmark_action is not None:
            benchmark_action.triggered.connect(self._open_gp_regression_widget)
        else:
            benchmarks_menu = self.menuBar().addMenu("Benchmarks")
            self._action_gp_regression = benchmarks_menu.addAction("GP Regression")
            self._action_gp_regression.triggered.connect(self._open_gp_regression_widget)

    def _open_gp_regression_widget(self) -> None:
        if self._benchmark_gp is None:
            QMessageBox.warning(
                self,
                "Benchmarks Unavailable",
                "GP benchmark interface could not be imported.",
            )
            return

        self._benchmark_gp.show()

    def _on_new_data_setup_clicked(self) -> None:
        if self.workspace_root is None:
            QMessageBox.information(
                self,
                "Workspace Required",
                "Select File -> New.. or File -> Load before creating a data setup.",
            )
            return

        dialog = NewDataSetupDialog(workspace_root=self.workspace_root, parent=self)
        if dialog.exec() != QDialog.Accepted:
            return

        if dialog.created_data_root is not None:
            self._broadcast_log(f"[DATA] Created data instance: {dialog.created_data_root}")
            self._set_data_selection_for_all_models(dialog.created_data_root)
            self._broadcast_log(f"[DATA] Active data instance for all models: {dialog.created_data_root}")

    def _on_new_workspace_clicked(self) -> None:
        if self._process_by_model:
            QMessageBox.information(
                self,
                "Workspace Busy",
                "Stop running setup/embedding jobs before switching workspace.",
            )
            return

        selected_dir = QFileDialog.getExistingDirectory(
            self,
            "Select new workspace directory",
            str(self.workspace_root or Path.home()),
        )
        if not selected_dir:
            return

        try:
            workspace = Path(selected_dir).resolve()
            workspace.mkdir(parents=True, exist_ok=True)
            (workspace / "models").mkdir(parents=True, exist_ok=True)
            marker = self._workspace_marker_path(workspace)
            if not marker.exists():
                marker.write_text(
                    json.dumps({"format": "morehub-workspace", "version": 1}, indent=2) + "\n",
                    encoding="utf-8",
                )
            self._set_workspace(workspace, persist=True)
        except Exception as exc:
            QMessageBox.critical(self, "Workspace Error", f"Failed to initialize workspace:\n{exc}")

    def _on_load_workspace_clicked(self) -> None:
        if self._process_by_model:
            QMessageBox.information(
                self,
                "Workspace Busy",
                "Stop running setup/embedding jobs before switching workspace.",
            )
            return

        selected_dir = QFileDialog.getExistingDirectory(
            self,
            "Load existing workspace directory",
            str(self.workspace_root or Path.home()),
        )
        if not selected_dir:
            return

        workspace = Path(selected_dir).resolve()
        if not self._is_workspace_valid(workspace):
            QMessageBox.warning(
                self,
                "Invalid Workspace",
                (
                    "Selected folder is not a valid workspace.\n\n"
                    "Expected either:\n"
                    "- a ./models directory, or\n"
                    f"- a {WORKSPACE_MARKER_NAME} marker file."
                ),
            )
            return

        self._set_workspace(workspace, persist=True)

    def _workspace_marker_path(self, workspace_root: Path) -> Path:
        return workspace_root / WORKSPACE_MARKER_NAME

    def _workspace_layout_paths(self, workspace_root: Path) -> dict[str, Path]:
        return {
            "models": workspace_root / "models",
        }

    def _ensure_workspace_layout(self, workspace_root: Path) -> None:
        for path in self._workspace_layout_paths(workspace_root).values():
            path.mkdir(parents=True, exist_ok=True)

    def _default_input_directory(self) -> Path:
        if self.workspace_root is None:
            return Path.home()
        return self.workspace_root

    def _is_workspace_valid(self, workspace_root: Path) -> bool:
        if not workspace_root.exists() or not workspace_root.is_dir():
            return False
        return (workspace_root / "models").is_dir() or self._workspace_marker_path(workspace_root).exists()

    def _set_workspace(self, workspace_root: Path, persist: bool) -> None:
        previous_workspace = self.workspace_root
        self.workspace_root = workspace_root.resolve()
        if previous_workspace != self.workspace_root:
            self._clear_all_data_selections()
        self._ensure_workspace_layout(self.workspace_root)
        marker = self._workspace_marker_path(self.workspace_root)
        if not marker.exists():
            marker.write_text(
                json.dumps({"format": "morehub-workspace", "version": 1}, indent=2) + "\n",
                encoding="utf-8",
            )
        if persist:
            self._persist_workspace(self.workspace_root)
        self._update_window_title()
        self._refresh_all_statuses()
        if self._benchmark_gp is not None:
            self._benchmark_gp.refresh_workspace_state()
        self._broadcast_log(f"[WORKSPACE] Active workspace: {self.workspace_root}")

    def _persist_workspace(self, workspace_root: Path) -> None:
        APP_STATE_DIR.mkdir(parents=True, exist_ok=True)
        payload = {"workspace_root": str(workspace_root)}
        APP_STATE_FILE.write_text(json.dumps(payload, indent=2) + "\n", encoding="utf-8")

    def _load_persisted_workspace(self) -> None:
        if not APP_STATE_FILE.exists():
            return
        try:
            payload = json.loads(APP_STATE_FILE.read_text(encoding="utf-8"))
            raw_workspace = payload.get("workspace_root")
            if not raw_workspace:
                return
            workspace = Path(raw_workspace).resolve()
            if self._is_workspace_valid(workspace):
                self._set_workspace(workspace, persist=False)
        except Exception:
            # Best-effort restore only.
            return

    def _update_window_title(self) -> None:
        if self.workspace_root is None:
            self.setWindowTitle("MoreHub - No workspace selected")
            return
        self.setWindowTitle(f"MoreHub - {self.workspace_root}")

    def _broadcast_log(self, text: str) -> None:
        for console in self._console_by_model.values():
            console.append(text.rstrip())

    def _set_model_data_label(self, model_name: str, data_root: Path | None) -> None:
        label = self._input_label_by_model.get(model_name)
        if label is None:
            return

        if data_root is None:
            label.setText("No data selected")
            label.setToolTip("")
            return

        label.setText(data_root.name)
        label.setToolTip(str(data_root))

    def _clear_all_data_selections(self) -> None:
        self._global_data_root = None
        self._selected_input_by_model.clear()
        for model_name in self._input_label_by_model:
            self._set_model_data_label(model_name, None)

    def _set_data_selection_for_all_models(self, data_root: Path) -> None:
        resolved = Path(data_root).resolve()
        self._global_data_root = resolved
        for model_name in self._input_label_by_model:
            self._selected_input_by_model[model_name] = resolved
            self._set_model_data_label(model_name, resolved)

    def _on_model_tab_changed(self, index: int) -> None:
        if index < 0:
            return
        model_name = self.ui.tabWidget.tabText(index).strip().upper()
        data_root = self._selected_input_by_model.get(model_name) or self._global_data_root
        if data_root is None:
            self._set_model_data_label(model_name, None)
            return
        self._set_model_data_label(model_name, Path(data_root).resolve())

    def _bind_run_buttons(self) -> None:
        run_button_names = [
            "pushButton_2",
            "pushButton_4",
            "pushButton_6",
            "pushButton_8",
            "pushButton_10",
            "pushButton_12",
            "pushButton_14",
            "pushButton_16",
            "pushButton_18",
            "pushButton_20",
            "pushButton_22",
            "pushButton_24",
        ]
        console_names = [
            "textEdit",
            "textEdit_2",
            "textEdit_3",
            "textEdit_4",
            "textEdit_5",
            "textEdit_6",
            "textEdit_7",
            "textEdit_8",
            "textEdit_9",
            "textEdit_10",
            "textEdit_11",
            "textEdit_12",
        ]

        for index, button_name in enumerate(run_button_names):
            button = getattr(self.ui, button_name, None)
            console = getattr(self.ui, console_names[index], None)
            if button is None or not isinstance(console, QTextEdit):
                continue

            tab_label = self.ui.tabWidget.tabText(index).strip()
            model_name = tab_label.upper()

            self._model_by_button[button] = model_name
            self._console_by_model[model_name] = console
            button.clicked.connect(self._on_run_clicked)
    def _bind_input_buttons(self) -> None:
        input_button_names = [
            "pushButton",
            "pushButton_3",
            "pushButton_5",
            "pushButton_7",
            "pushButton_9",
            "pushButton_11",
            "pushButton_13",
            "pushButton_15",
            "pushButton_17",
            "pushButton_19",
            "pushButton_21",
            "pushButton_23",
        ]
        input_label_names = [
            "label_2",
            "label_11",
            "label_23",
            "label_33",
            "label_43",
            "label_49",
            "label_29",
            "label_55",
            "label_59",
            "label_65",
            "label_71",
            "label_77",
        ]

        for index, button_name in enumerate(input_button_names):
            button = getattr(self.ui, button_name, None)
            label = getattr(self.ui, input_label_names[index], None)
            if not isinstance(button, QPushButton) or not isinstance(label, QLabel):
                continue

            model_name = self.ui.tabWidget.tabText(index).strip().upper()
            self._input_button_to_model[button] = model_name
            self._input_label_by_model[model_name] = label
            button.clicked.connect(self._on_input_select_clicked)

    def _on_input_select_clicked(self) -> None:
        button = self.sender()
        model_name = self._input_button_to_model.get(button)
        if not model_name:
            return

        selected_dir = QFileDialog.getExistingDirectory(
            self,
            f"Select data folder for {model_name}",
            str(self._default_input_directory()),
        )
        if not selected_dir:
            return

        data_root = Path(selected_dir).resolve()
        if not is_data_instance(data_root):
            inferred = find_data_root(data_root)
            if inferred is None or not is_data_instance(inferred):
                QMessageBox.warning(
                    self,
                    "Invalid Data Folder",
                    "Select a valid data instance folder (data_<name>) that contains .morehub_data.json.",
                )
                self._append_log(model_name, f"[{model_name}] Invalid data folder: {data_root}")
                return
            data_root = inferred

        self._set_data_selection_for_all_models(data_root)
        self._append_log(model_name, f"[{model_name}] Data selected for all models: {data_root}")
        try:
            molecules_path, _target_path, detected_format = resolve_data_source(data_root)
            self._append_log(model_name, f"[{model_name}] Data selected: {data_root}")
            self._append_log(model_name, f"[{model_name}] Source molecules: {molecules_path}")
            self._append_log(model_name, f"[{model_name}] Input format from source metadata: {detected_format}")
        except Exception:
            self._append_log(model_name, f"[{model_name}] Data selected: {data_root}")

    def _bind_format_controls(self) -> None:
        smiles_names = [
            "radioButton",
            "radioButton_3",
            "radioButton_5",
            "radioButton_7",
            "radioButton_9",
            "radioButton_11",
            "radioButton_13",
            "radioButton_18",
            "radioButton_20",
            "radioButton_24",
            "radioButton_30",
            "radioButton_35",
        ]
        selfies_names = [
            "radioButton_2",
            "radioButton_4",
            "radioButton_6",
            "radioButton_8",
            "radioButton_10",
            "radioButton_12",
            "radioButton_14",
            "radioButton_19",
            "radioButton_21",
            "radioButton_29",
            "radioButton_31",
            "radioButton_36",
        ]

        for index, smiles_name in enumerate(smiles_names):
            smiles_radio = getattr(self.ui, smiles_name, None)
            selfies_radio = getattr(self.ui, selfies_names[index], None)
            if not isinstance(smiles_radio, QRadioButton) or not isinstance(selfies_radio, QRadioButton):
                continue
            model_name = self.ui.tabWidget.tabText(index).strip().upper()
            self._format_widgets_by_model[model_name] = {
                "smiles": smiles_radio,
                "selfies": selfies_radio,
            }
            # Safe default for now while format toggle behavior is evolving.
            if not smiles_radio.isChecked() and not selfies_radio.isChecked():
                selfies_radio.setChecked(True)
    def _get_selected_input_format(self, model_name: str) -> str:
        radios = self._format_widgets_by_model.get(model_name)
        if not radios:
            return "selfies"
        if radios["smiles"].isChecked():
            return "smiles"
        return "selfies"

    def _bind_embeddings_option_controls(self) -> None:
        # Grover-specific controls from the third tab.
        use_large = getattr(self.ui, "checkBox_10", None)
        use_bonds = getattr(self.ui, "checkBox_9", None)
        use_atoms = getattr(self.ui, "checkBox_8", None)
        if (
            isinstance(use_large, QCheckBox)
            and isinstance(use_bonds, QCheckBox)
            and isinstance(use_atoms, QCheckBox)
        ):
            self._grover_option_widgets = {
                "use_large": use_large,
                "use_bonds": use_bonds,
                "use_atoms": use_atoms,
            }
            if not use_atoms.isChecked() and not use_bonds.isChecked():
                use_atoms.setChecked(True)
            self._embeddings_extra_args_builders["GROVER"] = self._build_grover_extra_args

        # ChemBERTa-specific checkpoint variant selection from the seventh tab.
        chemberta_zinc = getattr(self.ui, "radioButton_15", None)
        chemberta_mtr = getattr(self.ui, "radioButton_16", None)
        chemberta_mlm = getattr(self.ui, "radioButton_17", None)
        if (
            isinstance(chemberta_zinc, QRadioButton)
            and isinstance(chemberta_mtr, QRadioButton)
            and isinstance(chemberta_mlm, QRadioButton)
        ):
            self._chemberta_variant_widgets = {
                "zinc-base-v1": chemberta_zinc,
                "77m-mtr": chemberta_mtr,
                "77m-mlm": chemberta_mlm,
            }
            if not any(widget.isChecked() for widget in self._chemberta_variant_widgets.values()):
                chemberta_zinc.setChecked(True)
            self._embeddings_extra_args_builders["CHEMBERTA"] = self._build_chemberta_extra_args

        # NovoMolGen-specific checkpoint variant selection from the ninth tab.
        novomolgen_32m_atomwise = getattr(self.ui, "radioButton_22", None)
        novomolgen_157m_atomwise = getattr(self.ui, "radioButton_23", None)
        novomolgen_300m_atomwise = getattr(self.ui, "radioButton_25", None)
        novomolgen_32m_bpe = getattr(self.ui, "radioButton_26", None)
        novomolgen_157m_bpe = getattr(self.ui, "radioButton_27", None)
        novomolgen_300m_bpe = getattr(self.ui, "radioButton_28", None)
        if (
            isinstance(novomolgen_32m_atomwise, QRadioButton)
            and isinstance(novomolgen_157m_atomwise, QRadioButton)
            and isinstance(novomolgen_300m_atomwise, QRadioButton)
            and isinstance(novomolgen_32m_bpe, QRadioButton)
            and isinstance(novomolgen_157m_bpe, QRadioButton)
            and isinstance(novomolgen_300m_bpe, QRadioButton)
        ):
            self._novomolgen_variant_widgets = {
                "32m-atomwise": novomolgen_32m_atomwise,
                "157m-atomwise": novomolgen_157m_atomwise,
                "300m-atomwise": novomolgen_300m_atomwise,
                "32m-bpe": novomolgen_32m_bpe,
                "157m-bpe": novomolgen_157m_bpe,
                "300m-bpe": novomolgen_300m_bpe,
            }
            if not any(widget.isChecked() for widget in self._novomolgen_variant_widgets.values()):
                novomolgen_32m_atomwise.setChecked(True)
            self._embeddings_extra_args_builders["NOVOMOLGEN"] = self._build_novomolgen_extra_args

        # MolGen-specific checkpoint selection from the eleventh tab.
        molgen_large = getattr(self.ui, "radioButton_32", None)
        molgen_large_opt = getattr(self.ui, "radioButton_33", None)
        molgen_7b = getattr(self.ui, "radioButton_34", None)
        if (
            isinstance(molgen_large, QRadioButton)
            and isinstance(molgen_large_opt, QRadioButton)
            and isinstance(molgen_7b, QRadioButton)
        ):
            self._molgen_variant_widgets = {
                "large": molgen_large,
                "large-opt": molgen_large_opt,
                "7b": molgen_7b,
            }
            if not any(widget.isChecked() for widget in self._molgen_variant_widgets.values()):
                molgen_7b.setChecked(True)
            self._embeddings_extra_args_builders["MOLGEN"] = self._build_molgen_extra_args

        # MolCLR-specific backbone selection from the twelfth tab.
        molclr_gin = getattr(self.ui, "radioButton_37", None)
        molclr_gcn = getattr(self.ui, "radioButton_38", None)
        if molclr_gin is None or molclr_gcn is None:
            molclr_gin = getattr(self.ui, "checkBox_11", None)
            molclr_gcn = getattr(self.ui, "checkBox_12", None)

        if isinstance(molclr_gin, (QCheckBox, QRadioButton)) and isinstance(molclr_gcn, (QCheckBox, QRadioButton)):
            self._molclr_backbone_widgets = {
                "gin": molclr_gin,
                "gcn": molclr_gcn,
            }
            if not molclr_gin.isChecked() and not molclr_gcn.isChecked():
                molclr_gcn.setChecked(True)
            self._embeddings_extra_args_builders["MOLCLR"] = self._build_molclr_extra_args

        # Optional per-model overwrite toggles. If present in a tab and the label contains
        # "overwrite", that checkbox controls replacement of previous embeddings for that model.
        for index in range(self.ui.tabWidget.count()):
            tab = self.ui.tabWidget.widget(index)
            if tab is None:
                continue
            model_name = self.ui.tabWidget.tabText(index).strip().upper()
            for checkbox in tab.findChildren(QCheckBox):
                text = checkbox.text().strip().lower()
                if "overwrite" in text:
                    self._overwrite_widget_by_model[model_name] = checkbox
                    break
    def _build_model_extra_args(self, model_name: str) -> list[str]:
        builder = self._embeddings_extra_args_builders.get(model_name)
        if builder is None:
            return []
        return builder()

    def _extract_variant_arg(self, extra_args: list[str]) -> str | None:
        for index, token in enumerate(extra_args):
            if token == "--variant" and (index + 1) < len(extra_args):
                value = str(extra_args[index + 1]).strip()
                return value or None
        return None

    def _normalize_output_token(self, value: str) -> str:
        token = re.sub(r"[^A-Za-z0-9]+", "_", str(value)).strip("_")
        return token or "default"

    def _build_embeddings_version_tag(self, model_name: str, extra_args: list[str]) -> str | None:
        upper = model_name.upper()

        if upper == "GROVER":
            variant = self._extract_variant_arg(extra_args) or "base"
            use_atoms = "--use-atoms" in extra_args
            use_bonds = "--use-bonds" in extra_args
            if use_atoms and use_bonds:
                mode = "atomsBonds"
            elif use_atoms:
                mode = "atoms"
            elif use_bonds:
                mode = "bonds"
            else:
                mode = "atoms"
            return f"{variant}_{mode}"

        if upper == "NOVOMOLGEN":
            variant = (self._extract_variant_arg(extra_args) or "").lower()
            variant_map = {
                "32m-atomwise": "NovoMolGen_32M_SMILES_AtomWise",
                "157m-atomwise": "NovoMolGen_157M_SMILES_AtomWise",
                "300m-atomwise": "NovoMolGen_300M_SMILES_AtomWise",
                "32m-bpe": "NovoMolGen_32M_SMILES_BPE",
                "157m-bpe": "NovoMolGen_157M_SMILES_BPE",
                "300m-bpe": "NovoMolGen_300M_SMILES_BPE",
            }
            if variant in variant_map:
                return variant_map[variant]
            return variant or None

        variant = self._extract_variant_arg(extra_args)
        return variant or None

    def _build_embeddings_output_stem(self, model_name: str, extra_args: list[str]) -> tuple[str, str | None]:
        version_tag = self._build_embeddings_version_tag(model_name, extra_args)
        parts = ["embeddings", model_name.upper()]
        if version_tag:
            parts.append(version_tag)
        stem = "_".join(self._normalize_output_token(part) for part in parts if part)
        return stem, version_tag

    def _build_grover_extra_args(self) -> list[str]:
        widgets = self._grover_option_widgets
        if not widgets:
            return []

        use_large = widgets["use_large"].isChecked()
        use_bonds = widgets["use_bonds"].isChecked()
        use_atoms = widgets["use_atoms"].isChecked()

        args = ["--variant", "large" if use_large else "base"]
        if use_atoms:
            args.append("--use-atoms")
        if use_bonds:
            args.append("--use-bonds")
        if not use_atoms and not use_bonds:
            args.append("--use-atoms")
            self._append_log("GROVER", "[GROVER] Neither atoms nor bonds selected; defaulting to atoms.")
        return args

    def _build_chemberta_extra_args(self) -> list[str]:
        widgets = self._chemberta_variant_widgets
        if not widgets:
            return []

        for variant, radio in widgets.items():
            if radio.isChecked():
                return ["--variant", variant]

        # Defensive fallback for unexpected UI state.
        self._append_log("CHEMBERTA", "[CHEMBERTA] No variant selected; defaulting to zinc-base-v1.")
        return ["--variant", "zinc-base-v1"]

    def _build_novomolgen_extra_args(self) -> list[str]:
        widgets = self._novomolgen_variant_widgets
        if not widgets:
            return []

        for variant, radio in widgets.items():
            if radio.isChecked():
                return ["--variant", variant]

        # Defensive fallback for unexpected UI state.
        self._append_log("NOVOMOLGEN", "[NOVOMOLGEN] No variant selected; defaulting to 32m-atomwise.")
        return ["--variant", "32m-atomwise"]

    def _build_molgen_extra_args(self) -> list[str]:
        widgets = self._molgen_variant_widgets
        if not widgets:
            return []

        for variant, radio in widgets.items():
            if radio.isChecked():
                return ["--variant", variant]

        # Defensive fallback for unexpected UI state.
        self._append_log("MOLGEN", "[MOLGEN] No variant selected; defaulting to 7b.")
        return ["--variant", "7b"]

    def _build_molclr_extra_args(self) -> list[str]:
        widgets = self._molclr_backbone_widgets
        if not widgets:
            return []

        use_gin = widgets["gin"].isChecked()
        use_gcn = widgets["gcn"].isChecked()

        if use_gin and use_gcn:
            self._append_log("MOLCLR", "[MOLCLR] Both GIN and GCN selected; defaulting to GIN.")
            return ["--variant", "gin"]
        if use_gin:
            return ["--variant", "gin"]
        if use_gcn:
            return ["--variant", "gcn"]

        self._append_log("MOLCLR", "[MOLCLR] No backbone selected; defaulting to GCN.")
        return ["--variant", "gcn"]

    def _should_overwrite_embeddings(self, model_name: str) -> bool:
        widget = self._overwrite_widget_by_model.get(model_name)
        if widget is None:
            return False
        return bool(widget.isChecked())

    def _bind_status_widgets(self) -> None:
        model_installed_names = [
            "model_installed",
            "model_installed_2",
            "model_installed_3",
            "model_installed_4",
            "model_installed_5",
            "model_installed_6",
            "model_installed_7",
            "model_installed_8",
            "model_installed_9",
            "model_installed_10",
            "model_installed_11",
            "model_installed_12",
        ]
        checkpoints_names = [
            "checkpoints_downloaded",
            "checkpoints_downloaded_2",
            "checkpoints_downloaded_3",
            "checkpoints_downloaded_4",
            "checkpoints_downloaded_5",
            "checkpoints_downloaded_6",
            "checkpoints_downloaded_7",
            "checkpoints_downloaded_8",
            "checkpoints_downloaded_9",
            "checkpoints_downloaded_10",
            "checkpoints_downloaded_11",
            "checkpoints_downloaded_12",
        ]

        for index in range(len(model_installed_names)):
            model_checkbox = getattr(self.ui, model_installed_names[index], None)
            checkpoints_checkbox = getattr(self.ui, checkpoints_names[index], None)
            if not isinstance(model_checkbox, QCheckBox) or not isinstance(checkpoints_checkbox, QCheckBox):
                continue

            model_name = self.ui.tabWidget.tabText(index).strip().upper()
            self._status_widgets_by_model[model_name] = {
                "model_installed": model_checkbox,
                "checkpoints_downloaded": checkpoints_checkbox,
            }

            for checkbox in (model_checkbox, checkpoints_checkbox):
                checkbox.setEnabled(True)
                checkbox.setCheckable(True)
                checkbox.setFocusPolicy(Qt.NoFocus)
                checkbox.setAttribute(Qt.WA_TransparentForMouseEvents, True)
    def _refresh_all_statuses(self) -> None:
        for model_name in self._status_widgets_by_model:
            self._refresh_model_status(model_name)

    def _refresh_model_status(self, model_name: str) -> None:
        widgets = self._status_widgets_by_model.get(model_name)
        if not widgets:
            return
        model_installed, checkpoints_downloaded = self._get_model_state(model_name)
        widgets["model_installed"].setChecked(model_installed)
        widgets["checkpoints_downloaded"].setChecked(checkpoints_downloaded)

    def _load_model_metadata(self, model_name: str) -> dict[str, Any] | None:
        metadata_dir = self.metadata_root / model_name
        candidates = [
            metadata_dir / "model.json",
            metadata_dir / f"{metadata_dir.name}.json",
            metadata_dir / f"{model_name}.json",
        ]
        for path in candidates:
            if not path.exists():
                continue
            try:
                with path.open("r", encoding="utf-8") as handle:
                    return json.load(handle)
            except Exception:
                return None
        return None

    def _repository_artifacts_present(self, model_dir: Path, metadata: dict[str, Any]) -> bool | None:
        runtime_cfg = metadata.get("runtime") or {}
        runtime_source = str(runtime_cfg.get("source", "")).strip().lower()
        if not runtime_source:
            runtime_source = "repository" if metadata.get("repository") else "standalone_script"

        if runtime_source != "repository":
            return None

        repo_cfg = metadata.get("repository")
        if not isinstance(repo_cfg, dict) or not bool(repo_cfg.get("artifactsInRepository", False)):
            return None

        raw_artifact_paths = repo_cfg.get("artifactPaths") or []
        if isinstance(raw_artifact_paths, str):
            raw_artifact_paths = [raw_artifact_paths]

        model_dir_abs = model_dir.resolve()
        expected_paths = []
        for raw in raw_artifact_paths:
            if not isinstance(raw, str) or not raw.strip():
                continue
            candidate = (model_dir_abs / raw).resolve()
            try:
                candidate.relative_to(model_dir_abs)
            except ValueError:
                continue
            expected_paths.append(candidate)

        if expected_paths:
            return all(path.exists() for path in expected_paths)

        repo_subdir = str(repo_cfg.get("dir", "")).strip()
        if repo_subdir:
            return (model_dir_abs / repo_subdir).exists()
        return (model_dir_abs / ".repo").exists()

    def _get_model_state(self, model_name: str) -> tuple[bool, bool]:
        if self.workspace_root is None:
            return False, False

        model_dir = self.workspace_root / "models" / model_name
        sentinel = model_dir / ".setup_complete"
        venv_dir = model_dir / ".venv"
        checkpoints_dir_data = model_dir / "data"
        checkpoints_dir_cache = model_dir / ".cache"

        model_installed = venv_dir.exists() and sentinel.exists()

        checkpoints_downloaded: bool
        metadata = self._load_model_metadata(model_name)
        if metadata is None:
            checkpoints_downloaded = checkpoints_dir_data.exists() or checkpoints_dir_cache.exists()
        else:
            repo_artifacts_present = self._repository_artifacts_present(model_dir, metadata)
            if repo_artifacts_present is None:
                checkpoints_downloaded = checkpoints_dir_data.exists() or checkpoints_dir_cache.exists()
            else:
                checkpoints_downloaded = repo_artifacts_present

        return model_installed, checkpoints_downloaded

    def _on_run_clicked(self) -> None:
        button = self.sender()  # sender() returns the button that was clicked
        model_name = self._model_by_button.get(button)
        if not model_name:
            return

        if self.workspace_root is None:
            self._append_log(model_name, "[WORKSPACE] Select File -> New.. or File -> Load before running.")
            QMessageBox.information(
                self,
                "Workspace Required",
                "Select File -> New.. or File -> Load before running a model.",
            )
            return

        model_installed, checkpoints_downloaded = self._get_model_state(model_name)

        if model_installed and checkpoints_downloaded:
            self.run_model_embeddings(model_name)
            return

        if model_installed and not checkpoints_downloaded:
            self._append_log(
                model_name,
                f"[{model_name}] Model is installed but checkpoints are missing. Downloading checkpoints only...",
            )
            self.run_model_setup(model_name, download_only=True, run_embeddings_after=True)
            return

        if not model_installed and checkpoints_downloaded:
            self._append_log(
                model_name,
                f"[{model_name}] Checkpoints exist but model is not installed. Installing environment...",
            )
            self.run_model_setup(model_name, download_only=False, run_embeddings_after=True)
            return

        self._append_log(model_name, f"[{model_name}] Model and checkpoints missing. Running full setup...")
        self.run_model_setup(model_name, download_only=False, run_embeddings_after=True)

    def _log_color_for_line(self, line: str) -> str | None:
        normalized = line.strip().lower()
        if "setup finished successfully" in normalized or "embeddings finished:" in normalized:
            return "#1b8f3a"
        return None

    def _append_log(self, model_name: str, text: str) -> None:
        console = self._console_by_model.get(model_name)
        if console is None:
            return

        stripped = text.rstrip()
        for line in stripped.splitlines() or [stripped]:
            safe_line = html.escape(line)
            color = self._log_color_for_line(line) or "#000000"
            console.append(f"<span style=\"color: {color};\">{safe_line}</span>")

    def _sha256_file(self, path: Path) -> str:
        digest = hashlib.sha256()
        with path.open("rb") as handle:
            for chunk in iter(lambda: handle.read(1024 * 1024), b""):
                digest.update(chunk)
        return digest.hexdigest()

    def _capture_embeddings_output(self, model_name: str, process: QProcess, stream_name: str) -> None:
        if stream_name == "stdout":
            raw = bytes(process.readAllStandardOutput())
        else:
            raw = bytes(process.readAllStandardError())

        text = raw.decode(errors="replace")
        if not text:
            return

        self._append_log(model_name, text)
        run_context = self._active_embeddings_runs.get(model_name)
        if run_context is not None:
            run_context["log_chunks"].append(f"[{stream_name}] {text}")

    def _finalize_embeddings_run(self, model_name: str, exit_code: int, output_path: Path) -> None:
        run_context = self._active_embeddings_runs.pop(model_name, None)
        if run_context is None:
            return

        run_dir = run_context["run_dir"]
        log_path = run_context["log_path"]
        manifest_path = run_context["manifest_path"]

        try:
            log_path.write_text("".join(run_context["log_chunks"]), encoding="utf-8")
        except Exception as exc:
            self._append_log(model_name, f"[{model_name}] Warning: could not write run log ({exc})")

        output_files = []
        if run_dir.exists():
            for item in sorted(run_dir.iterdir()):
                if item.is_file():
                    output_files.append(item.name)

        manifest = {
            "tool": "embeddings",
            "model": model_name,
            "model_version": run_context.get("model_version"),
            "run_id": run_context["run_id"],
            "started_at": run_context["started_at"],
            "finished_at": datetime.now(timezone.utc).isoformat(),
            "exit_code": exit_code,
            "status": "success" if exit_code == 0 else "failed",
            "workspace_root": run_context.get("workspace_root") or "",
            "data_root": run_context.get("data_root") or "",
            "input": {
                "path": run_context["input_path"],
                "sha256": run_context["input_sha256"],
                "format": run_context["input_format"],
                "targets_path": run_context.get("source_targets_path"),
            },
            "extra_args": run_context["extra_args"],
            "command": run_context["command"],
            "output": {
                "stem": run_context.get("output_stem"),
                "path": str(output_path),
                "exists": output_path.exists(),
                "files": output_files,
            },
            "log_file": str(log_path),
        }

        try:
            manifest_path.write_text(json.dumps(manifest, indent=2) + "\n", encoding="utf-8")
            self._append_log(model_name, f"[{model_name}] Run manifest: {manifest_path}")
        except Exception as exc:
            self._append_log(model_name, f"[{model_name}] Warning: could not write run manifest ({exc})")


    def run_model_setup(
        self,
        model_name: str,
        download_only: bool = False,
        run_embeddings_after: bool = False,
    ) -> None:
        if model_name in self._process_by_model:
            self._append_log(model_name, f"[{model_name}] Setup is already running.")
            return
        if self.workspace_root is None:
            self._append_log(model_name, "[WORKSPACE] No workspace selected.")
            return

        metadata_dir = self.metadata_root / model_name
        setup_script = metadata_dir / "setup.py"

        if not setup_script.exists():
            self._append_log(model_name, f"[{model_name}] setup.py not found at {setup_script}")
            return

        process = QProcess(self)
        process.setWorkingDirectory(str(metadata_dir))
        process.setProgram(sys.executable)
        args = [str(setup_script), "--model", model_name, "--workspace", str(self.workspace_root)]
        if download_only:
            args.append("--download-only")
        process.setArguments(args)
        process.setProcessChannelMode(QProcess.SeparateChannels)

        process.readyReadStandardOutput.connect(
            lambda m=model_name, p=process: self._append_log(
                m, bytes(p.readAllStandardOutput()).decode(errors="replace")
            )
        )
        process.readyReadStandardError.connect(
            lambda m=model_name, p=process: self._append_log(
                m, bytes(p.readAllStandardError()).decode(errors="replace")
            )
        )
        process.finished.connect(
            lambda exit_code, _status, m=model_name: self._on_setup_finished(m, exit_code)
        )

        self._process_by_model[model_name] = process
        self._run_embeddings_after_setup[model_name] = run_embeddings_after
        phase = "download-only setup" if download_only else "setup"
        self._append_log(model_name, f"[{model_name}] Starting {phase}...")
        process.start()

    def run_model_embeddings(self, model_name: str) -> None:
        if model_name in self._process_by_model:
            self._append_log(model_name, f"[{model_name}] Another process is already running.")
            return
        if self.workspace_root is None:
            self._append_log(model_name, "[WORKSPACE] No workspace selected.")
            return

        data_root = self._selected_input_by_model.get(model_name) or self._global_data_root
        if data_root is None:
            self._append_log(model_name, f"[{model_name}] No data folder selected.")
            self._append_log(model_name, "[DATA] Select once using any folder button; selection is shared across all models.")
            return
        data_root = Path(data_root).resolve()
        if not is_data_instance(data_root):
            inferred = find_data_root(data_root)
            if inferred is None or not is_data_instance(inferred):
                self._append_log(model_name, f"[{model_name}] Invalid data folder: {data_root}")
                return
            data_root = inferred
            self._set_data_selection_for_all_models(data_root)

        try:
            molecules_path, target_path, source_input_format = resolve_data_source(data_root)
        except Exception as exc:
            self._append_log(model_name, f"[{model_name}] Failed to resolve data source: {exc}")
            return

        radio_input_format = self._get_selected_input_format(model_name)
        input_format = source_input_format
        if radio_input_format != source_input_format:
            self._append_log(
                model_name,
                (
                    f"[{model_name}] UI format '{radio_input_format}' differs from source metadata "
                    f"'{source_input_format}'. Using source metadata format."
                ),
            )

        extra_args = self._build_model_extra_args(model_name)
        output_stem, model_version = self._build_embeddings_output_stem(model_name, extra_args)
        model_output_base = model_embeddings_dir(data_root, model_name)
        if model_version:
            variant_dir = self._normalize_output_token(model_version).lower()
            model_output_root = model_output_base / variant_dir
        else:
            model_output_root = model_output_base

        if model_output_root.exists():
            shutil.rmtree(model_output_root, ignore_errors=True)
            self._append_log(model_name, f"[{model_name}] Overwriting previous embeddings in: {model_output_root}")

        model_output_root.mkdir(parents=True, exist_ok=True)
        output_path = model_output_root / f"{output_stem}.npy"

        try:
            program, args, cwd, output_path, run_dir, _run_id = build_embeddings_command(
                workspace_root=self.workspace_root,
                model_name=model_name,
                input_file=molecules_path,
                input_format=input_format,
                output_file=output_path,
                extra_args=extra_args,
                metadata_root=self.metadata_root,
            )
        except FileNotFoundError as exc:
            self._append_log(model_name, f"[{model_name}] {exc}")
            return
        except Exception as exc:
            self._append_log(model_name, f"[{model_name}] Failed to build embeddings command: {exc}")
            return

        run_context = {
            "run_id": "latest",
            "model_version": model_version,
            "output_stem": output_stem,
            "run_dir": run_dir,
            "manifest_path": run_dir / "manifest.json",
            "log_path": run_dir / "stdout.log",
            "started_at": datetime.now(timezone.utc).isoformat(),
            "workspace_root": str(self.workspace_root),
            "data_root": str(data_root),
            "input_path": str(molecules_path.resolve()),
            "input_sha256": self._sha256_file(molecules_path),
            "input_format": input_format,
            "source_targets_path": str(target_path.resolve()) if target_path is not None else None,
            "extra_args": list(extra_args),
            "command": {
                "program": program,
                "args": args,
                "cwd": cwd,
            },
            "log_chunks": [],
        }
        self._active_embeddings_runs[model_name] = run_context

        process = QProcess(self)
        process.setWorkingDirectory(cwd)
        process.setProgram(program)
        process.setArguments(args)
        process.setProcessChannelMode(QProcess.SeparateChannels)

        process.readyReadStandardOutput.connect(
            lambda m=model_name, p=process: self._capture_embeddings_output(m, p, "stdout")
        )
        process.readyReadStandardError.connect(
            lambda m=model_name, p=process: self._capture_embeddings_output(m, p, "stderr")
        )
        process.finished.connect(
            lambda exit_code, _status, m=model_name, out=output_path: self._on_embeddings_finished(
                m, exit_code, out
            )
        )

        self._process_by_model[model_name] = process
        self._append_log(model_name, f"[{model_name}] Starting embeddings...")
        self._append_log(model_name, f"[{model_name}] Data root: {data_root}")
        self._append_log(model_name, f"[{model_name}] Input format: {input_format}")
        if extra_args:
            self._append_log(model_name, f"[{model_name}] Extra args: {' '.join(extra_args)}")
        self._append_log(model_name, f"[{model_name}] Output: {output_path}")
        process.start()

    def _on_setup_finished(self, model_name: str, exit_code: int) -> None:
        # Clean up the process reference and check if we need to run embeddings next
        self._process_by_model.pop(model_name, None)
        run_embeddings_after = self._run_embeddings_after_setup.pop(model_name, False)
        if exit_code == 0:
            self._append_log(model_name, f"[{model_name}] Setup finished successfully.")
        else:
            self._append_log(model_name, f"[{model_name}] Setup failed with exit code {exit_code}.")
        self._refresh_model_status(model_name)
        if exit_code == 0 and run_embeddings_after:
            self.run_model_embeddings(model_name)

    def _on_embeddings_finished(self, model_name: str, exit_code: int, output_path: Path) -> None:
        self._process_by_model.pop(model_name, None)
        if exit_code == 0:
            self._append_log(model_name, f"[{model_name}] Embeddings finished: {output_path}")
        else:
            self._append_log(model_name, f"[{model_name}] Embeddings failed with exit code {exit_code}.")
        self._finalize_embeddings_run(model_name, exit_code, output_path)



class NewDataSetupDialog(QDialog):
    def __init__(self, workspace_root: Path, parent=None) -> None:
        super().__init__(parent)
        self.ui = Ui_NewDataForm()
        self.ui.setupUi(self)

        self.workspace_root = workspace_root
        self.created_data_root: Path | None = None
        self._molecules_path: Path | None = None
        self._targets_path: Path | None = None

        self.setWindowTitle("New Data Setup")
        self._configure_icons(parent)

        self.ui.pushButton.clicked.connect(self._on_pick_molecules)
        self.ui.pushButton_3.clicked.connect(self._on_pick_targets)
        self.ui.pushButton_2.clicked.connect(self._on_process)

    def _configure_icons(self, parent) -> None:
        package_root = getattr(parent, "package_root", Path(__file__).resolve().parent)
        icons_dir = Path(package_root) / "ui" / "icons"
        folder_icon = QIcon(str(icons_dir / "folder.png"))
        run_icon = QIcon(str(icons_dir / "run.png"))

        self.ui.pushButton.setIcon(folder_icon)
        self.ui.pushButton_3.setIcon(folder_icon)
        self.ui.pushButton_2.setIcon(run_icon)

    def _on_pick_molecules(self) -> None:
        selected, _ = QFileDialog.getOpenFileName(
            self,
            "Select molecules file (selfies/smiles)",
            str(self.workspace_root),
            "Text files (*.txt);;All files (*.*)",
        )
        if not selected:
            return
        self._molecules_path = Path(selected).resolve()
        self.ui.label_2.setText(self._molecules_path.name)
        self.ui.label_2.setToolTip(str(self._molecules_path))

    def _on_pick_targets(self) -> None:
        selected, _ = QFileDialog.getOpenFileName(
            self,
            "Select targets file (optional)",
            str(self.workspace_root),
            "Text/CSV files (*.txt *.csv);;All files (*.*)",
        )
        if not selected:
            return
        self._targets_path = Path(selected).resolve()
        self.ui.label_6.setText(self._targets_path.name)
        self.ui.label_6.setToolTip(str(self._targets_path))

    def _on_process(self) -> None:
        if self._molecules_path is None:
            QMessageBox.warning(self, "Data Setup", "Select a molecules file first.")
            return

        data_name = self.ui.lineEdit.text().strip()
        if not data_name:
            QMessageBox.warning(self, "Data Setup", "Provide a name for your data.")
            return

        copy_only = bool(self.ui.checkBox.isChecked())

        try:
            self.created_data_root = create_data_instance(
                workspace_root=self.workspace_root,
                data_name=data_name,
                molecules_path=self._molecules_path,
                target_path=self._targets_path,
                copy_only=copy_only,
            )
        except Exception as exc:
            QMessageBox.critical(self, "Data Setup", f"Failed to create data setup:\n{exc}")
            return

        QMessageBox.information(self, "Data Setup", f"Created data instance:\n{self.created_data_root}")
        self.accept()


def main() -> None:
    qt_app = QApplication(sys.argv)
    qt_app.setStyle("Fusion")
    window = App()
    window.show()
    sys.exit(qt_app.exec())


if __name__ == "__main__":
    main()







import argparse
import os
import sys
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import torch
import torch.nn.functional as F
from rdkit import Chem
from rdkit.Chem.rdchem import BondType as BT

if not str(torch.__version__).startswith("1.7.1"):
    raise RuntimeError(
        "MOLCLR requires torch==1.7.1+cpu. "
        "Found torch=%s. Recreate the MOLCLR .venv and run setup again." % torch.__version__
    )

try:
    from torch_geometric.data import Data, DataLoader
except Exception as exc:
    raise ImportError(
        "Could not import torch_geometric with the installed torch extension stack. "
        "Ensure torch==1.7.1+cpu and matching torch-scatter/torch-sparse wheels are installed."
    ) from exc


SCRIPT_DIR = Path(__file__).resolve().parent
REPO_DIR = SCRIPT_DIR / ".repo" / "MOLCLR"

if str(REPO_DIR) not in sys.path:
    sys.path.insert(0, str(REPO_DIR))

try:
    from models.gcn_molclr import GCN
    from models.ginet_molclr import GINet
except Exception as exc:
    raise ImportError(
        "Could not import MolCLR modules from the cloned repository. "
        "Ensure MOLCLR setup completed successfully."
    ) from exc

PROJECT_ROOT = Path(__file__).resolve().parent.parent.parent
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))

from embeddings_preprocess import (
    load_input_lines,
    preprocess_lines,
    resolve_output_paths,
    write_failures_csv,
    write_mapping_csv,
)

ATOM_LIST = list(range(1, 119))
CHIRALITY_LIST = [
    Chem.rdchem.ChiralType.CHI_UNSPECIFIED,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
    Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
]
BOND_LIST = [BT.SINGLE, BT.DOUBLE, BT.TRIPLE, BT.AROMATIC]
BONDDIR_LIST = [
    Chem.rdchem.BondDir.NONE,
    Chem.rdchem.BondDir.ENDUPRIGHT,
    Chem.rdchem.BondDir.ENDDOWNRIGHT,
]


def _safe_index(value, choices, default_idx=0):
    try:
        return choices.index(value)
    except ValueError:
        return default_idx


def _resolve_device(device_arg):
    normalized = str(device_arg).strip().lower()
    if normalized == "auto":
        return torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if normalized not in {"cpu", "cuda"}:
        raise ValueError("Unsupported device: %s" % device_arg)
    if normalized == "cuda" and not torch.cuda.is_available():
        print("CUDA requested but unavailable; falling back to CPU.", file=sys.stderr)
        return torch.device("cpu")
    return torch.device(normalized)


def _normalize_variant(variant):
    normalized = str(variant).strip().lower()
    if normalized not in {"gin", "gcn"}:
        raise ValueError("Unknown MolCLR variant '%s'. Expected gin or gcn." % variant)
    return normalized


def mol_to_graph_data_obj_simple(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        raise ValueError("Invalid SMILES: %s" % smiles)

    type_idx = []
    chirality_idx = []
    for atom in mol.GetAtoms():
        atom_num = atom.GetAtomicNum()
        if atom_num not in ATOM_LIST:
            raise ValueError("Unsupported atomic number %s for MolCLR featurization." % atom_num)
        type_idx.append(ATOM_LIST.index(atom_num))
        chirality_idx.append(_safe_index(atom.GetChiralTag(), CHIRALITY_LIST, default_idx=0))

    x = torch.tensor(list(zip(type_idx, chirality_idx)), dtype=torch.long)

    row = []
    col = []
    edge_feat = []
    for bond in mol.GetBonds():
        start = bond.GetBeginAtomIdx()
        end = bond.GetEndAtomIdx()
        bond_type_idx = _safe_index(bond.GetBondType(), BOND_LIST, default_idx=0)
        bond_dir_idx = _safe_index(bond.GetBondDir(), BONDDIR_LIST, default_idx=0)

        row += [start, end]
        col += [end, start]
        edge_feat += [[bond_type_idx, bond_dir_idx], [bond_type_idx, bond_dir_idx]]

    if row:
        edge_index = torch.tensor([row, col], dtype=torch.long)
        edge_attr = torch.tensor(edge_feat, dtype=torch.long)
    else:
        edge_index = torch.empty((2, 0), dtype=torch.long)
        edge_attr = torch.empty((0, 2), dtype=torch.long)

    return Data(x=x, edge_index=edge_index, edge_attr=edge_attr)


def _build_model_and_checkpoint(variant, feat_dim):
    if variant == "gin":
        model = GINet(num_layer=5, emb_dim=300, feat_dim=feat_dim, drop_ratio=0, pool="mean")
        ckpt_subdir = "pretrained_gin"
    elif variant == "gcn":
        model = GCN(num_layer=5, emb_dim=300, feat_dim=feat_dim, drop_ratio=0, pool="mean")
        ckpt_subdir = "pretrained_gcn"
    else:
        raise ValueError("Unknown MolCLR variant '%s'." % variant)

    checkpoint_path = REPO_DIR / "ckpt" / ckpt_subdir / "checkpoints" / "model.pth"
    return model, checkpoint_path


def _load_checkpoint(model, checkpoint_path, device):
    if not checkpoint_path.exists():
        raise FileNotFoundError(
            "MolCLR checkpoint not found at %s. Ensure repository checkpoints are present."
            % checkpoint_path
        )

    state_obj = torch.load(str(checkpoint_path), map_location=device)
    if isinstance(state_obj, dict) and "state_dict" in state_obj:
        state_obj = state_obj["state_dict"]

    if isinstance(state_obj, dict):
        normalized = {}
        for key, value in state_obj.items():
            if key.startswith("module."):
                normalized[key[len("module.") :]] = value
            else:
                normalized[key] = value
        state_obj = normalized

    model.load_state_dict(state_obj)


def _encode_smiles(
    smiles_list,
    variant,
    feat_dim,
    embedding_type,
    batch_size,
    device,
    normalize,
):
    model, checkpoint_path = _build_model_and_checkpoint(variant=variant, feat_dim=feat_dim)
    model = model.to(device)
    _load_checkpoint(model, checkpoint_path, device)
    model.eval()

    graphs = []
    for smiles in smiles_list:
        graphs.append(mol_to_graph_data_obj_simple(smiles))

    loader = DataLoader(graphs, batch_size=batch_size, shuffle=False)

    chunks = []
    with torch.no_grad():
        for batch in loader:
            batch = batch.to(device)
            representation, projection = model(batch)
            if embedding_type == "representation":
                emb = representation
            else:
                emb = projection
            if normalize:
                emb = F.normalize(emb, dim=-1)
            chunks.append(emb.detach().cpu().numpy().astype(np.float32, copy=False))

    if not chunks:
        return np.empty((0, 0), dtype=np.float32)
    return np.concatenate(chunks, axis=0)


def main():
    parser = argparse.ArgumentParser(description="Encode molecules with MolCLR and save embeddings")
    parser.add_argument("--file", required=True, help="Path to input .txt file")
    parser.add_argument(
        "--input-format",
        default="selfies",
        choices=["selfies", "smiles"],
        help="Format of rows in input file",
    )
    parser.add_argument(
        "--model-input-type",
        default="smiles",
        choices=["selfies", "smiles"],
        help="Expected model input type from metadata",
    )
    parser.add_argument("--output-dir", default=".", help="Directory to write outputs")
    parser.add_argument("--output-stem", default="embeddings", help="Base output filename stem")
    parser.add_argument(
        "--variant",
        default="gcn",
        choices=["gin", "gcn"],
        help="MolCLR backbone checkpoint",
    )
    parser.add_argument(
        "--embedding",
        default="projected",
        choices=["projected", "representation"],
        help="Use projection head output or encoder representation",
    )
    parser.add_argument("--feat-dim", type=int, default=512)
    parser.add_argument("--batch-size", type=int, default=128)
    parser.add_argument(
        "--device",
        default="auto",
        choices=["auto", "cpu", "cuda"],
        help="Inference device",
    )
    parser.add_argument(
        "--normalize",
        action="store_true",
        help="Apply L2 normalization to each output embedding",
    )
    args = parser.parse_args()

    variant = _normalize_variant(args.variant)
    device = _resolve_device(args.device)

    input_path = Path(args.file).resolve()
    output_dir = Path(args.output_dir).resolve()
    npy_path, map_path, failures_path = resolve_output_paths(output_dir, args.output_stem)

    lines = load_input_lines(input_path)
    records, failures = preprocess_lines(
        lines,
        input_format=args.input_format,
        model_input_type=args.model_input_type,
    )

    valid_records = []
    smiles_inputs = []
    for record in records:
        smiles_text = record["smiles"]
        try:
            _ = mol_to_graph_data_obj_simple(smiles_text)
        except Exception as exc:
            failures.append(
                {
                    "original_line": str(record.get("original_line", "")),
                    "input_text": record.get("input_text", ""),
                    "reason": "molclr_featurization_failed:%s" % exc,
                }
            )
            continue
        valid_records.append(record)
        smiles_inputs.append(smiles_text)

    if smiles_inputs:
        embeddings = _encode_smiles(
            smiles_inputs,
            variant=variant,
            feat_dim=args.feat_dim,
            embedding_type=args.embedding,
            batch_size=args.batch_size,
            device=device,
            normalize=args.normalize,
        )
    else:
        embeddings = np.empty((0, 0), dtype=np.float32)

    np.save(str(npy_path), embeddings)
    write_mapping_csv(map_path, valid_records)
    write_failures_csv(failures_path, failures)

    print("Saved embeddings to %s with shape %s." % (npy_path, embeddings.shape))
    print("Saved row mapping to %s." % map_path)
    if failures:
        print("Saved %d failed rows to %s." % (len(failures), failures_path))
    print("Model source: %s (variant=%s)" % (REPO_DIR, variant))


if __name__ == "__main__":
    main()
